# -*- coding: utf-8 -*-
"""
Module to prepare the training datasets.
"""

from __future__ import print_function
import logging
import os
import shutil
import math
import datetime
from pathlib import Path
from typing import Optional, Tuple

import fiona
import pandas as pd
import geopandas as gpd
import numpy as np
import owslib
import rasterio as rio
import rasterio.features as rio_features
import shapely.geometry as sh_geom

from orthoseg.helpers import log_helper
from orthoseg.util import ows_util
from orthoseg.util import geofile_util

#-------------------------------------------------------------
# First define/init some general variables/constants
#-------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)
#logger.setLevel(logging.DEBUG)

#-------------------------------------------------------------
# The real work
#-------------------------------------------------------------

def prepare_traindatasets(
        label_files: dict,
        classes: dict,
        image_layers: dict,
        training_dir: Path,
        image_pixel_x_size: float = 0.25,
        image_pixel_y_size: float = 0.25,
        image_pixel_width: int = 512,
        image_pixel_height: int = 512,
        max_samples: int = 5000,
        output_filelist_csv: str = '',
        output_keep_nested_dirs: bool = False,
        force: bool = False) -> Tuple[Path, int]:
    """
    This function prepares training data for the vector labels provided.

    It will:
        * get orthophoto's from a WMS server
        * create the corresponding label mask for each orthophoto
        
    Returns a tuple with (output_dir, dataversion):
            output_dir: the dir where the traindataset was created/found
            dataversion: a version number for the dataset created/found

    Args
        label_files: paths to the files with label data and locations to generate images for
        wms_server_url: WMS server where the images can be fetched from
        wms_layername: layername on the WMS server to use
        output_basedir: the base dir where the train dataset needs to be written to 

    """
    image_srs_width = math.fabs(image_pixel_width*image_pixel_x_size)   # tile width in units of crs => 500 m
    image_srs_height = math.fabs(image_pixel_height*image_pixel_y_size) # tile height in units of crs => 500 m

    # Determine the current data version based on existing output data dir(s),
    # but ignore dirs ending on _ERROR
    output_dirs = training_dir.glob(f"[0-9]*/")
    output_dirs = [output_dir for output_dir in output_dirs if not '_BUSY' in output_dir.name]
    logger.info(f"output_dirs: {output_dirs}")
    if len(output_dirs) == 0:
        output_legacy_train_dirs = training_dir.glob(f"train_[0-9]*/")
        output_legacy_train_dirs = [output_dir for output_dir in output_legacy_train_dirs if output_dir.name.endswith('_BUSY') is False]
        if len(output_legacy_train_dirs) == 0:
            dataversion_new = 1
        else:
            # Get the output dir with the highest version (=first if sorted desc)
            output_legacy_dir_mostrecent = sorted(output_legacy_train_dirs, reverse=True)[0]
            dataversion_mostrecent = int(output_legacy_dir_mostrecent.name.split('_')[1])
            dataversion_new = dataversion_mostrecent + 1
    else:
        # Get the output dir with the highest version (=first if sorted desc)
        output_dir_mostrecent = sorted(output_dirs, reverse=True)[0]
        dataversion_mostrecent = int(output_dir_mostrecent.name)
        
        # If none of the input files changed since previous run, reuse dataset
        reuse = False
        for label_file_key in label_files:                  
            label_file = label_files[label_file_key]
            reuse = True
            labellocations_output_mostrecent_path = (
                    output_dir_mostrecent / label_file['locations_path'].name)
            labeldata_output_mostrecent_path = output_dir_mostrecent / label_file['data_path'].name
            if(not (labellocations_output_mostrecent_path.exists()
               and labeldata_output_mostrecent_path.exists()
               and geofile_util.cmp(
                        label_file['locations_path'], labellocations_output_mostrecent_path)
               and geofile_util.cmp(label_file['data_path'], labeldata_output_mostrecent_path))):
                logger.info(f"RETURN: input label file(s) changed since last prepare_traindatasets, recreate")
                reuse = False
                break
        if reuse == True:
            return (output_dir_mostrecent, dataversion_mostrecent)
        else:
            dataversion_new = dataversion_mostrecent + 1
    
    # Prepare the output basedir...
    for i in range(100):
        output_tmp_basedir = training_dir / f"{dataversion_new:02d}_BUSY_{i:02d}"
        if output_tmp_basedir.exists():
            try:
                shutil.rmtree(output_tmp_basedir)
            except:
                None
        if not output_tmp_basedir.exists():
            try:
                output_tmp_basedir.mkdir(parents=True)
                break
            except:
                None

    # Process all input files
    labellocations_gdf = None
    labeldata_gdf = None
    logger.info(label_files)
    for label_file_key in label_files:
        label_file = label_files[label_file_key]

        # Copy the vector files to the dest dir so we keep knowing which files 
        # were used to create the dataset
        geofile_util.copy(label_file['locations_path'], output_tmp_basedir)
        geofile_util.copy(label_file['data_path'], output_tmp_basedir)

        # Read label data and append to general dataframes
        logger.debug(f"Read label locations from {label_file['locations_path']}")
        file_labellocations_gdf = geofile_util.read_file(label_file['locations_path'])
        if file_labellocations_gdf is not None and len(file_labellocations_gdf) > 0:
            file_labellocations_gdf.loc[:, 'image_layer'] = label_file['image_layer']
            if labellocations_gdf is None:
                labellocations_gdf = file_labellocations_gdf
            else:
                labellocations_gdf = gpd.GeoDataFrame(
                        pd.concat([labellocations_gdf, file_labellocations_gdf], ignore_index=True),
                        crs=file_labellocations_gdf.crs)
        else:
            logger.warn(f"No label locations data found in {label_file['locations_path']}")
        logger.debug(f"Read label data from {label_file['data_path']}")
        file_labeldata_gdf = geofile_util.read_file(label_file['data_path'])
        if file_labeldata_gdf is not None and len(file_labeldata_gdf) > 0:
            file_labeldata_gdf.loc[:, 'image_layer'] = label_file['image_layer']
            if labeldata_gdf is None:
                labeldata_gdf = file_labeldata_gdf
            else:
                labeldata_gdf = gpd.GeoDataFrame(
                        pd.concat([labeldata_gdf, file_labeldata_gdf], ignore_index=True),
                        crs=file_labeldata_gdf.crs)
        else:
            logger.warn(f"No label data found in {label_file['data_path']}")

    # Get the srs to use from the input vectors...
    try:
        img_srs = labellocations_gdf.crs['init']
    except Exception as ex:
        logger.exception(f"Error getting crs from labellocations, labellocations_gdf.crs: {labellocations_gdf.crs}")
        raise ex
    
    # Create list with only the input labels that need to be burned in the mask
    if labeldata_gdf is not None and 'label_name' in labeldata_gdf.columns:
        # If there is a column 'label_name', filter on the labels provided
        labels_to_burn_gdf = (
                labeldata_gdf.loc[labeldata_gdf['label_name'].isin(classes)]).copy()
        labels_to_burn_gdf['burn_value'] = 0
        for label_name in classes:
            labels_to_burn_gdf.loc[(labels_to_burn_gdf['label_name'] == label_name),
                                   'burn_value'] = classes[label_name]['burn_value']
        if len(labeldata_gdf) != len(labels_to_burn_gdf):
            logger.warn(f"Number of labels to burn changed from {len(labeldata_gdf)} to {len(labels_to_burn_gdf)} with filter on classes: {classes}")
    elif len(classes) == 2:
        labels_to_burn_gdf = labeldata_gdf
        labels_to_burn_gdf.loc[:, 'burn_value'] = classes[list(classes)[1]]['burn_value']
    else:
        raise Exception(f"Column 'label_name' is mandatory in labeldata if multiple classes specified: {classes}")
                    
    # Prepare the different traindata types
    for traindata_type in ['train', 'validation', 'test']:
                   
        # Create the output dir's if they don't exist yet...
        output_tmp_dir = output_tmp_basedir / traindata_type
        output_tmp_image_dir = output_tmp_dir / 'image'
        output_tmp_mask_dir = output_tmp_dir / 'mask'

        # Create output dirs...
        for dir in [output_tmp_dir, output_tmp_mask_dir, output_tmp_image_dir]:
            if dir and not dir.exists():
                dir.mkdir(parents=True)

        try:
            # Get the label locations for this traindata type
            labels_to_use_for_bounds_gdf = (
                    labellocations_gdf[labellocations_gdf['traindata_type'] == traindata_type])
            
            # Loop trough all locations labels to get an image for each of them
            nb_todo = len(labels_to_use_for_bounds_gdf)
            nb_processed = 0
            logger.info(f"Get images for {nb_todo} {traindata_type} labels")
            created_images_gdf = gpd.GeoDataFrame()
            created_images_gdf['geometry'] = None
            start_time = datetime.datetime.now()
            wms_servers = {}
            for i, label_tuple in enumerate(labels_to_use_for_bounds_gdf.itertuples()):
                
                # TODO: update the polygon if it doesn't match the size of the image...
                # as a start... ideally make sure we use it entirely for cases with
                # no abundance of data
                # Make sure the image requested is of the correct size
                label_geom = label_tuple.geometry            
                geom_bounds = label_geom.bounds
                xmin = geom_bounds[0]-(geom_bounds[0]%image_pixel_x_size)
                ymin = geom_bounds[1]-(geom_bounds[1]%image_pixel_y_size)
                xmax = xmin + image_srs_width
                ymax = ymin + image_srs_height
                img_bbox = sh_geom.box(xmin, ymin, xmax, ymax)
                image_layer = getattr(label_tuple, 'image_layer')

                # If the wms to be used hasn't been initialised yet
                if image_layer not in wms_servers:
                    wms_servers[image_layer] = owslib.wms.WebMapService(
                            url=image_layers[image_layer]['wms_server_url'], 
                            version=image_layers[image_layer]['wms_version'])
                                                
                # Now really get the image
                logger.debug(f"Get image for coordinates {img_bbox.bounds}")
                image_filepath = ows_util.getmap_to_file(
                        wms=wms_servers[image_layer],
                        layers=image_layers[image_layer]['wms_layernames'],
                        styles=image_layers[image_layer]['wms_layerstyles'],
                        output_dir=output_tmp_image_dir,
                        srs=img_srs,
                        bbox=img_bbox.bounds,
                        size=(image_pixel_width, image_pixel_height),
                        image_format=ows_util.FORMAT_PNG,
                        #image_format_save=ows_util.FORMAT_TIFF,
                        image_pixels_ignore_border=image_layers[image_layer]['image_pixels_ignore_border'],
                        transparent=False,
                        layername_in_filename=True)

                # Create a mask corresponding with the image file
                # image_filepath can be None if file existed already, so check if not None...
                if image_filepath is not None:
                    # Mask should not be in a lossy format!
                    mask_filepath = Path(str(image_filepath)
                            .replace(str(output_tmp_image_dir), str(output_tmp_mask_dir))
                            .replace('.jpg', '.png'))
                    nb_classes = len(classes)
                    # Only keep the labels that are meant for this image layer
                    labels_gdf = (labels_to_burn_gdf.loc[
                                        labels_to_burn_gdf['image_layer'] == image_layer]).copy()
                    if len(labels_gdf) == 0:
                        logger.info("No labels to be burned for this layer, this is weird!")
                    _create_mask(
                            input_image_filepath=image_filepath,
                            output_mask_filepath=mask_filepath,
                            labels_to_burn_gdf=labels_gdf,
                            nb_classes=nb_classes,
                            force=force)
                
                # Log the progress and prediction speed
                nb_processed += 1
                time_passed = (datetime.datetime.now()-start_time).total_seconds()
                if time_passed > 0 and nb_processed > 0:
                    processed_per_hour = (nb_processed/time_passed) * 3600
                    hours_to_go = (int)((nb_todo - i)/processed_per_hour)
                    min_to_go = (int)((((nb_todo - i)/processed_per_hour)%1)*60)
                    print(f"\r{hours_to_go:3d}:{min_to_go:2d} left for {nb_todo-i} of {nb_todo} at {processed_per_hour:0.0f}/h", 
                          end="", flush=True)
        except Exception as ex:
            message = "Error preparing dataset!"
            raise Exception(message) from ex

    # If everything went fine, rename output_tmp_dir to the final output_dir
    output_basedir = training_dir / f"{dataversion_new:02d}"
    os.rename(output_tmp_basedir, output_basedir)

    return (output_basedir, dataversion_new)

''' Not maintained!
def create_masks_for_images(
        input_vector_label_filepath: str,
        input_image_dir: str,
        output_basedir: str,
        image_subdir: str = 'image',
        mask_subdir: str = 'mask',
        burn_value: int = 255,
        force: bool = False):

    # Check if the input file exists, if not, return
    if not os.path.exists(input_vector_label_filepath):
        message = f"Input file doesn't exist, so do nothing and return: {input_vector_label_filepath}"
        raise Exception(message)
    # Check if the input file exists, if not, return
    if not os.path.exists(input_image_dir):
        message = f"Input image dir doesn't exist, so do nothing and return: {input_image_dir}"
        raise Exception(message)
    
    # Determine the current data version based on existing output data dir(s), but ignore dirs ending on _ERROR
    output_dirs = glob.glob(f"{output_basedir}_*")
    output_dirs = [output_dir for output_dir in output_dirs if output_dir.endswith('_BUSY') is False]
    if len(output_dirs) == 0:
        dataversion_new = 1
    else:
        # Get the output dir with the highest version (=first if sorted desc)
        output_dir_mostrecent = sorted(output_dirs, reverse=True)[0]
        output_subdir_mostrecent = os.path.basename(output_dir_mostrecent)
        dataversion_mostrecent = int(output_subdir_mostrecent.split('_')[1])
        dataversion_new = dataversion_mostrecent + 1
        
        # If the input vector label file didn't change since previous run 
        # dataset can be reused
        output_vector_mostrecent_filepath = os.path.join(
                output_dir_mostrecent, os.path.basename(input_vector_label_filepath))
        if(os.path.exists(output_vector_mostrecent_filepath)
           and geofile_util.cmp(input_vector_label_filepath, 
                                  output_vector_mostrecent_filepath)):
            logger.info(f"RETURN: input vector label file isn't changed since last prepare_traindatasets, so no need to recreate")
            return output_dir_mostrecent, dataversion_mostrecent

    # Create the output dir's if they don't exist yet...
    output_dir = f"{output_basedir}_{dataversion_new:02d}"
    output_tmp_dir = f"{output_basedir}_{dataversion_new:02d}_BUSY"
    output_tmp_image_dir = os.path.join(output_tmp_dir, image_subdir)
    output_tmp_mask_dir = os.path.join(output_tmp_dir, mask_subdir)

    # Prepare the output dir...
    if os.path.exists(output_tmp_dir):
        shutil.rmtree(output_tmp_dir)
    for dir in [output_tmp_dir, output_tmp_mask_dir, output_tmp_image_dir]:
        if dir and not os.path.exists(dir):
            os.makedirs(dir)

    # Copy the vector file(s) to the dest dir so we keep knowing which file was
    # used to create the dataset
    geofile_util.copy(input_vector_label_filepath, output_tmp_dir)
    
    # Open vector layer
    logger.debug(f"Open vector file {input_vector_label_filepath}")
    input_label_gdf = gpd.read_file(input_vector_label_filepath)

    # Create list with only the input labels that are positive examples, as 
    # are the only ones that will need to be burned in the mask
    labels_to_burn_gdf = input_label_gdf[input_label_gdf['burninmask'] == 1]
    
    # Loop trough input images
    input_image_filepaths = glob.glob(f"{input_image_dir}/*.tif")
    logger.info(f"process {len(input_image_filepaths)} input images")
    for input_image_filepath in input_image_filepaths:
        _, input_image_filename = os.path.split(input_image_filepath)
        
        image_filepath = os.path.join(output_tmp_image_dir, input_image_filename)
        shutil.copyfile(input_image_filepath, image_filepath)

        mask_filepath = os.path.join(output_tmp_mask_dir, input_image_filename)
        _create_mask(
                input_image_filepath=image_filepath,
                output_mask_filepath=mask_filepath,
                labels_to_burn_gdf=labels_to_burn_gdf,
                force=force)
'''

def _create_mask(
        input_image_filepath: Path,
        output_mask_filepath: Path,
        labels_to_burn_gdf: gpd.geodataframe,
        nb_classes: int = 1,
        output_imagecopy_filepath: Optional[Path] = None,
        minimum_pct_labeled: float = 0.0,
        force: bool = False) -> Optional[bool]:

    # If file exists already and force is False... stop.
    if(force is False
       and output_mask_filepath.exists()):
        logger.debug(f"Output file already exist, and force is False, return: {output_mask_filepath}")
        return

    # Create a mask corresponding with the image file
    # First read the properties of the input image to copy them for the output
    logger.debug(f"Create mask to {output_mask_filepath}")
    with rio.open(input_image_filepath) as image_ds:
        image_input_profile = image_ds.profile
        image_transform_affine = image_ds.transform

    # Prepare the file profile for the mask depending on output type
    output_ext_lower = output_mask_filepath.suffix.lower()
    if output_ext_lower == '.tif':
        image_output_profile = rio.profiles.DefaultGTiffProfile(
                count=1, transform=image_transform_affine, crs=image_input_profile['crs'])
    if output_ext_lower == '.png':
        image_output_profile = rio.profiles.Profile(driver='PNG', count=1)
    else:
        raise Exception(f"Unsupported mask extension (should be a lossless format!): {output_ext_lower}")
    image_output_profile.update(
            width=image_input_profile['width'], height=image_input_profile['height'], 
            dtype=rio.uint8)

    # Burn the vectors in a mask
    burn_shapes = ((geom, value) 
            for geom, value in zip(labels_to_burn_gdf.geometry, labels_to_burn_gdf.burn_value) if geom is not None)
    try:
        mask_arr = rio_features.rasterize(
                shapes=burn_shapes, transform=image_transform_affine,
                dtype=rio.uint8, fill=0, 
                out_shape=(image_output_profile['width'], image_output_profile['height']))
    except Exception as ex:
        raise Exception(f"Error creating mask for {image_transform_affine}") from ex

    # Check if the mask meets the requirements to be written...
    if minimum_pct_labeled > 0:
        nb_pixels = np.size(mask_arr, 0) * np.size(mask_arr, 1)
        nb_pixels_data = nb_pixels - np.sum(mask_arr == 0)  #np.count_nonnan(image == NoData_value)
        logger.debug(f"nb_pixels: {nb_pixels}, nb_pixels_data: {nb_pixels_data}, pct data: {nb_pixels_data / nb_pixels}")

        if (nb_pixels_data / nb_pixels < minimum_pct_labeled):
            return False

    # Write the labeled mask
    image_output_profile = ows_util.get_cleaned_write_profile(image_output_profile)
    with rio.open(str(output_mask_filepath), 'w', **image_output_profile) as mask_ds:
        mask_ds.write(mask_arr, 1)
        
    return True

if __name__ == "__main__":
    raise Exception('Not implemented!')



# -*- coding: utf-8 -*-
"""
Process the tasks as configured in a tasks csv file.
"""

import argparse
import configparser
import datetime
from email.message import EmailMessage
import json
import logging
import logging.config
from pathlib import Path
import smtplib
from typing import List

# Because orthoseg isn't installed as package + it is higher in dir hierarchy, add root to sys.path
import sys
sys.path.insert(0, '.')

import pandas as pd

import orthoseg.util.git_downloader as git_downloader

def main():
    
    ##### Interprete arguments #####
    parser = argparse.ArgumentParser(add_help=False)

    # Optional arguments
    optional = parser.add_argument_group('Optional arguments')
    optional.add_argument('-t', '--tasksfile',
            help='The path to the tasks .csv file to use.')
    optional.add_argument('-s', '--sampleprojects', action='store_true',
            help="Download the sample projects to ~/orthoseg/sample_projects if this directory doesn't exist and run.")
    # Add back help         
    optional.add_argument('-h', '--help', action='help', default=argparse.SUPPRESS,
            help='Show this help message and exit')
    args = parser.parse_args()

    # If tasks file is specified, use it
    if args.tasksfile is not None:
        tasks_path = Path(args.tasksfile)
    elif args.sampleprojects is True:
        # If sampleprojects is true, download sampleproject and run
        orthoseg_dir = Path.home() / 'orthoseg'
        projects_dir = orthoseg_dir / 'sample_projects'
        if not projects_dir.exists():
            print("Download sample projects")
            git_repo_dir = 'https://github.com/theroggy/orthoseg/tree/master/sample_projects'
            git_downloader.download(repo_url=git_repo_dir, output_dir=orthoseg_dir)
        tasks_path = projects_dir / 'tasks.csv'
    else:
        parser.print_help()
        sys.exit(1)
    
    # Find the config filepaths to use
    config_filepaths = []
    script_dir = Path(__file__).resolve().parent
    config_defaults_path = script_dir / 'project_defaults.ini'
    if config_defaults_path.exists():
        config_filepaths.append(config_defaults_path)
    else:
        print(f"Warning: default project settings not found: {config_defaults_path}")
    config_defaults_overrule_path = tasks_path.parent / 'project_defaults_overrule.ini'
    if config_defaults_overrule_path.exists():
        config_filepaths.append(config_defaults_overrule_path)
    else:
        print(f"Warning: default overule project settings not found: {config_defaults_overrule_path}")

    # Run!
    run_tasks(
            tasks_path=tasks_path,
            config_filepaths=config_filepaths)

def run_tasks(
        tasks_path: Path,
        config_filepaths: List[Path],
        stop_on_error: bool = False):

    ##### Init #####
    # Read the configuration
    global runner_config
    runner_config = configparser.ConfigParser(
            interpolation=configparser.ExtendedInterpolation(),
            converters={'list': lambda x: [i.strip() for i in x.split(',')],
                        'listint': lambda x: [int(i.strip()) for i in x.split(',')],
                        'dict': lambda x: json.loads(x),
                        'path': lambda x: Path(x)})
    runner_config.read(config_filepaths)

    # Init logging
    logconfig_dict = runner_config['logging'].getdict('logconfig')

    # If there are file handlers, replace possible placeholders + make sure log dir exists
    for handler in logconfig_dict['handlers']:
        if "filename" in logconfig_dict['handlers'][handler]:
            # Format the filename
            log_path = Path(logconfig_dict['handlers'][handler]['filename'].format(
                    iso_datetime=f"{datetime.datetime.now():%Y-%m-%d_%H-%M-%S}"))

            # If the log_path is a relative path, resolve it towards the location of
            # the tasks file.
            if not log_path.is_absolute():
                log_path = tasks_path.parent / log_path
                print(f"Parameter logconfig.handlers.{handler}.filename was relative, so is now resolved to {log_path}")

            logconfig_dict['handlers'][handler]['filename'] = log_path.as_posix()

            # Also make sure the log dir exists
            log_path.parent.mkdir(parents=True, exist_ok=True)

    # Now load the log config
    logging.config.dictConfig(logconfig_dict)
    global logger
    logger = logging.getLogger()
    logger.info(f"Config files used for orthoseg: {config_filepaths}")
    
    # Read the tasks that need to be ran in the run_tasks file
    tasks_df = get_tasks(tasks_path)

    # Loop over tasks to run
    for task in tasks_df.itertuples(): 
        # If the task is not active, skip
        if(task.active == 0):
            continue

        # Format the projectfile_path
        projectfile_path = Path(task.config)
        # If the projectfile path is not absolute, treat is as relative to the tasks csv file...
        if not projectfile_path.is_absolute():
            projectfile_path = tasks_path.parent / projectfile_path

        # Now we are ready to start the action        
        logger.info(f"Start action {task.action} for config {projectfile_path}")
        try:
            if(task.action == 'train'):
                from orthoseg import train
                train.train(projectconfig_path=projectfile_path)
            elif(task.action == 'predict'):
                from orthoseg import predict
                predict.predict(projectconfig_path=projectfile_path)
            elif(task.action == 'load_images'):
                from orthoseg import load_images
                load_images.load_images(projectconfig_path=projectfile_path)
            elif(task.action == 'load_testsample_images'):
                from orthoseg import load_images
                load_images.load_images(
                        projectconfig_path=projectfile_path, 
                        load_testsample_images=True)
            elif(task.action == 'postprocess'):
                from orthoseg import postprocess
                postprocess.postprocess(projectconfig_path=projectfile_path)
            else:
                raise Exception(f"Unsupported action: {task.action}")

            message = f"Completed action {task.action} for config {projectfile_path}"
            logger.info(message)
            sendmail(message)
        except Exception as ex:
            message = f"ERROR in task with action {task.action} for config {projectfile_path}"
            logger.exception(message)
            sendmail(subject=message, body=f"Exception: {ex}")
            if stop_on_error:
                raise Exception(message) from ex

def get_tasks(filepath: Path):
    
    # Read
    tasks_df = pd.read_csv(filepath)

    # Trim column names and string columns
    for column in tasks_df.columns:
        column_stripped = column.strip()
        if column != column_stripped:
            tasks_df.rename(columns={column:column_stripped}, inplace=True)
        if tasks_df[column_stripped].dtype in ('str', 'object'):
            tasks_df[column_stripped] = tasks_df[column_stripped].astype(str).str.strip()

    # Check mandatory columns
    mandatory_columns = ['active', 'config', 'action']
    missing_columns = set(mandatory_columns).difference(set(tasks_df.columns))
    if(len(missing_columns) > 0):
        raise Exception(f"Missing column(s) in {filepath}: {missing_columns}")
    
    return tasks_df

def sendmail(
        subject: str, 
        body: str = None,
        stop_on_error: bool = False):

    if not runner_config['email'].getboolean('enabled', fallback=False):
        return
    mail_from = runner_config['email'].get('from', None)
    mail_to = runner_config['email'].get('to', None)
    mail_server = runner_config['email'].get('server', None)
    mail_server_username = runner_config['email'].get('username', None)
    mail_server_password = runner_config['email'].get('password', None)

    # If one of the necessary parameters not provided, log subject
    if(mail_from is None
       or mail_to is None
       or mail_server is None):
        logger.warning(f"Mail global_config not provided to send email with subject: {subject}")
        return

    try:
        # Create message
        msg = EmailMessage()
        msg.add_header('from', mail_from)
        msg.add_header('to', mail_to)
        msg.add_header('subject', subject)
        if body is not None:
            msg.set_payload(body)

        # Send the email
        server = smtplib.SMTP(mail_server)
        if(mail_server_username is not None
           and mail_server_password is not None):
            server.login(mail_server_username, mail_server_password)
        server.send_message(msg)
        server.quit()
    except Exception as ex:
        if stop_on_error is False:
            logger.exception("Error sending email")
        else:
            raise Exception("Error sending email") from ex

if __name__ == '__main__':
    main()



'''
Visvalingam-Whyatt method of poly-line vertex reduction

Visvalingam, M and Whyatt J D (1993)
"Line Generalisation by Repeated Elimination of Points", Cartographic J., 30 (1), 46 - 51

Described here:
http://web.archive.org/web/20100428020453/http://www2.dcs.hull.ac.uk/CISRG/publications/DPs/DP10/DP10.html

=========================================

The MIT License (MIT)

Copyright (c) 2014 Elliot Hallmark

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

================================
'''

from numpy import array, argmin
import numpy as np


def triangle_area(p1,p2,p3):
    """
    calculates the area of a triangle given its vertices
    """
    return abs(p1[0]*(p2[1]-p3[1])+p2[0]*(p3[1]-p1[1])+p3[0]*(p1[1]-p2[1]))/2.

def triangle_areas_from_array(arr):
    '''
    take an (N,2) array of points and return an (N,1)
    array of the areas of those triangles, where the first
    and last areas are np.inf

    see triangle_area for algorithm
    '''

    result = np.empty((len(arr),),arr.dtype)
    result[0] = np.inf; result[-1] = np.inf

    p1 = arr[:-2]
    p2 = arr[1:-1]
    p3 = arr[2:]
    
    #an accumulators to avoid unnecessary intermediate arrays
    accr = result[1:-1] #Accumulate directly into result
    acc1 = np.empty_like(accr)

    np.subtract(p2[:,1], p3[:,1], out = accr)
    np.multiply(p1[:,0], accr,    out = accr)
    np.subtract(p3[:,1], p1[:,1], out = acc1  )
    np.multiply(p2[:,0], acc1,    out = acc1  )
    np.add(acc1, accr,            out = accr)
    np.subtract(p1[:,1], p2[:,1], out = acc1  )
    np.multiply(p3[:,0], acc1,    out = acc1  )
    np.add(acc1, accr,            out = accr)
    np.abs(accr, out = accr)
    accr /= 2.
    #Notice: accr was writing into result, so the answer is in there
    return result

#the final value in thresholds is np.inf, which will never be
# the min value.  So, I am safe in "deleting" an index by
# just shifting the array over on top of it
def remove(s,i):
    '''
    Quick trick to remove an item from a numpy array without
    creating a new object.  Rather than the array shape changing,
    the final value just gets repeated to fill the space.

    ~3.5x faster than numpy.delete
    '''
    s[i:-1]=s[i+1:]

class VWSimplifier(object):

    def __init__(self,pts):
        '''Initialize with points. takes some time to build 
        the thresholds but then all threshold filtering later 
        is ultra fast'''
        self.pts = np.array(pts)
        self.thresholds = self.build_thresholds()
        self.ordered_thresholds = sorted(self.thresholds,reverse=True)

    def build_thresholds(self):
        '''compute the area value of each vertex, which one would
        use to mask an array of points for any threshold value.

        returns a numpy.array (length of pts)  of the areas.
        '''
        pts = self.pts
        nmax = len(pts)
        real_areas = triangle_areas_from_array(pts)
        real_indices = list(range(nmax))


        #destructable copies 
        #ARG! areas=real_areas[:] doesn't make a copy!
        areas = np.copy(real_areas)
        i = real_indices[:]
        
        #pick first point and set up for loop
        min_vert = argmin(areas)
        this_area = areas[min_vert]
        #  areas and i are modified for each point finished
        remove(areas,min_vert)   #faster
        #areas = np.delete(areas,min_vert) #slower
        real_idx = i.pop(min_vert)

        #cntr = 3
        while this_area<np.inf:
           '''min_vert was removed from areas and i.  Now,
           adjust the adjacent areas and remove the new 
           min_vert.

           Now that min_vert was filtered out, min_vert points 
           to the point after the deleted point.'''
           
           skip = False  #modified area may be the next minvert
           
           try:
             right_area = triangle_area(pts[i[min_vert-1]],
                            pts[i[min_vert]],pts[i[min_vert+1]])
           except IndexError:
             #trying to update area of endpoint. Don't do it
             pass
           else:
             right_idx = i[min_vert]
             if right_area <= this_area:
                 #even if the point now has a smaller area,
                 # it ultimately is not more significant than
                 # the last point, which needs to be removed
                 # first to justify removing this point.
                 # Though this point is the next most significant
                 right_area = this_area

                 #min_vert refers to the point to the right of 
                 # the previous min_vert, so we can leave it
                 # unchanged if it is still the min_vert
                 skip = min_vert

             #update both collections of areas
             real_areas[right_idx] = right_area
             areas[min_vert] = right_area
           
           if min_vert > 1:
             #cant try/except because 0-1=-1 is a valid index
             left_area = triangle_area(pts[i[min_vert-2]],
                           pts[i[min_vert-1]],pts[i[min_vert]])
             if left_area <= this_area:
                 #same justification as above
                 left_area = this_area
                 skip = min_vert-1
             real_areas[i[min_vert-1]] = left_area
             areas[min_vert-1] = left_area


           #only argmin if we have too.
           min_vert = skip or argmin(areas)
           real_idx = i.pop(min_vert)
           this_area = areas[min_vert]
           #areas = np.delete(areas,min_vert) #slower
           remove(areas,min_vert)  #faster
           '''if sum(np.where(areas==np.inf)[0]) != sum(list(reversed(range(len(areas))))[:cntr]):
             print "broke:",np.where(areas==np.inf)[0],cntr
             break
           cntr+=1
           #if real_areas[0]<np.inf or real_areas[-1]<np.inf:
           #  print "NO!", real_areas[0], real_areas[-1]
           '''
        return real_areas

    def from_threshold(self,threshold):
        return self.pts[self.thresholds >= threshold]

    def from_number(self,n):
        thresholds = self.ordered_thresholds
        try:
          threshold = thresholds[int(n)]
        except IndexError:
          return self.pts
        return self.pts[self.thresholds > threshold]

    def from_ratio(self,r):
        if r<=0 or r>1:
          raise ValueError("Ratio must be 0<r<=1")
        else:
          return self.from_number(r*len(self.thresholds))

class WKTSimplifier(VWSimplifier):
      '''VWSimplifier that returns strings suitable for WKT
      creation'''
      def __init__(self,*args,**kwargs):
         if 'precision' in kwargs:
           p=kwargs.pop('precision')
         else:
           p=None
         VWSimplifier.__init__(self,*args,**kwargs)
         self.set_precision(p)

      def set_precision(self,precision):
          if precision:
            self.pts_as_strs = self.pts.astype('S%s'%precision)
          else:
            self.pts_as_strs = self.pts.astype(str)


      '''slow
      def from_threshold(self,threshold,precision=None):
          arr = np.array2string(self.pts[self.thresholds > threshold],precision=precision)
          return arr.replace('[[ ','(').replace(']]',')').replace(']\n [ ',',')
      '''
      def wkt_from_threshold(self,threshold, precision=None):
          if precision:
            self.set_precision(precision)
          pts = self.pts_as_strs[self.thresholds >= threshold]
          return '(%s)'%','.join(['%s %s'%(x,y) for x,y in pts])

      def wkt_from_number(self,n,precision=None):
        thresholds = self.ordered_thresholds
        if n<3: n=3  #For polygons. TODO something better
        try:
          threshold = thresholds[int(n)]
        except IndexError:
          threshold = 0
  
        return self.wkt_from_threshold(threshold,precision=precision)

      def wkt_from_ratio(self,r,precision=None):
        if r<=0 or r>1:
          raise ValueError("Ratio must be 0<r<=1")
        else:
          return self.wkt_from_number(r*len(self.thresholds))



try:
    from django.contrib.gis.gdal import OGRGeometry,OGRException
    from django.contrib.gis.geos import GEOSGeometry, fromstr
except ImportError:         
    class GDALSimplifier(object):
        '''Dummy object that would be replaced by a real one if
           correct module exists'''
        def __init__(*args,**kwargs):
            print("""
                  django.contrib.gis.gdal not found.
                  GDALSimplifier not available.
                  """)
else:
    from json import loads
    import re
    p = re.compile( '([ 0123456789.]+) ([0123456789.]+)')
    class GDALSimplifier(object):
      '''Warning, there is a slight loss of precision just in the
      conversion from geometry object to numpy.array even if no
      threshold is applied.  ie:
      
      originalpolygeom.area   ->   413962.65495176613
      gdalsimplifierpoly.area ->   413962.65495339036
      '''
      def __init__(self,geom,precision=None,return_GDAL = True):
          '''accepts a gdal.OGRGeometry or geos.GEOSGeometry
          object and wraps multiple
          VWSimplifiers.  set return_GDAL to False for faster
          filtering with arrays of floats returned instead of
          geometry objects.'''
          global p
          self.return_GDAL = return_GDAL
          if isinstance(geom,OGRGeometry):
            name = geom.geom_name
            self.Geometry = lambda w: OGRGeometry(w,srs=geom.srs)
            self.pts = np.array(geom.tuple)
          elif isinstance(geom,GEOSGeometry):
            name = geom.geom_type.upper()
            self.Geometry = lambda w: fromstr(w) 
            self.pts = np.array(geom.tuple)
          elif isinstance(geom, unicode) or isinstance(geom,str): 
            #assume wkt
            #for WKT
            def str2tuple(q):
                return '(%s,%s)' % (q.group(1),q.group(2))

            self.return_GDAL = False #don't even try
            self.Geometry = lambda w: w #this will never be used
            name, pts = geom.split(' ',1)
            self.pts = loads(p.sub(str2tuple,pts).\
                             replace('(','[').replace(')',']'))
          self.precision = precision
          if name == 'LINESTRING':
            self.maskfunc = self.linemask
            self.buildfunc = self.linebuild
            self.fromnumfunc = self.notimplemented
          elif name == "POLYGON":
            self.maskfunc = self.polymask
            self.buildfunc = self.polybuild
            self.fromnumfunc = self.notimplemented
          elif name == "MULTIPOLYGON":
            self.maskfunc = self.multimask
            self.buildfunc = self.multibuild
            self.fromnumfunc = self.notimplemented
          else:
            raise OGRGeometryError("""
             Only types LINESTRING, POLYGON and MULTIPOLYGON
             supported, but got %s"""%name)
          #sets self.simplifiers to a list of VWSimplifiers
          self.buildfunc()

      #rather than concise, I'd rather be explicit and clear.
      
      def pt2str(self,pt):
           '''make length 2 numpy.array.__str__() fit for wkt'''
           return ' '.join(pt)

      def linebuild(self):
          self.simplifiers = [WKTSimplifier(self.pts)]

      def line2wkt(self,pts):
          return u'LINESTRING %s'%pts

      def linemask(self,threshold):
          get_pts = self.get_pts
          pts = get_pts(self.simplifiers[0],threshold)
          if self.return_GDAL:
            return self.Geometry(self.line2wkt(pts))
          else:
            return pts

      def polybuild(self):
          list_of_pts = self.pts
          result = []
          for pts in list_of_pts:
            result.append(WKTSimplifier(pts))
          self.simplifiers = result

      def poly2wkt(self,list_of_pts):
          return u'POLYGON (%s)'%','.join(list_of_pts)

      def polymask(self,threshold):
          get_pts = self.get_pts
          sims = self.simplifiers
          list_of_pts = [get_pts(sim,threshold) for sim in sims]
          if self.return_GDAL:
            return self.Geometry(self.poly2wkt(list_of_pts))
          else:
            return array(list_of_pts)

      def multibuild(self):
          list_of_list_of_pts = self.pts
          result = []
          for list_of_pts in list_of_list_of_pts:
            subresult = []
            for pts in list_of_pts:
              subresult.append(WKTSimplifier(pts))          
            result.append(subresult)
          self.simplifiers = result

      def multi2wkt(self,list_of_list_of_pts):
          outerlist = []
          for list_of_pts in list_of_list_of_pts:
            outerlist.append('(%s)'%','.join(list_of_pts))
          return u'MULTIPOLYGON (%s)'%','.join(outerlist)

      def multimask(self,threshold):
          loflofsims = self.simplifiers
          result = []
          get_pts = self.get_pts
          if self.return_GDAL:
            ret_func = lambda r: self.Geometry(self.multi2wkt(r))
          else:
            ret_func = lambda r: r
          for list_of_simplifiers in loflofsims:
            subresult = []
            for simplifier in list_of_simplifiers:
              subresult.append(get_pts(simplifier,threshold))
            result.append(subresult)
          return ret_func(result)

      def notimplemented(self,n):
          print("This function is not yet implemented")

      def from_threshold(self,threshold):
          precision = self.precision
          if self.return_GDAL:
            self.get_pts = lambda obj,t: obj.wkt_from_threshold(t,precision)
          else:
            self.get_pts = lambda obj,t: obj.from_threshold(t)
          return self.maskfunc(threshold)

      def from_number(self,n):
          precision = self.precision
          if self.return_GDAL:
            self.get_pts = lambda obj,t: obj.wkt_from_number(t,precision)
          else:
            self.get_pts = lambda obj,t: obj.from_number(t)
          return self.maskfunc(n)

      def from_ratio(self,r):
          precision = self.precision
          if self.return_GDAL:
            self.get_pts = lambda obj,t: obj.wkt_from_ratio(t,precision)
          else:
            self.get_pts = lambda obj,t: obj.from_ratio(t)
          return self.maskfunc(r)


def fancy_parametric(k):
    ''' good k's: .33,.5,.65,.7,1.3,1.4,1.9,3,4,5'''
    cos = np.cos
    sin = np.sin
    xt = lambda t: (k-1)*cos(t) + cos(t*(k-1))
    yt = lambda t: (k-1)*sin(t) - sin(t*(k-1))
    return xt,yt

if __name__ == "__main__":

   from time import time
   n = 5000
   thetas = np.linspace(0,16*np.pi,n)
   xt,yt = fancy_parametric(1.4)  
   pts = np.array([[xt(t),yt(t)] for t in thetas])
   start = time()
   simplifier = VWSimplifier(pts)
   pts = simplifier.from_number(1000)
   end = time()
   print("%s vertices removed in %02f seconds"%(n-len(pts), end-start))
   
   import matplotlib
   matplotlib.use('AGG')
   import matplotlib.pyplot as plot
   plot.plot(pts[:,0],pts[:,1],color='r')
   plot.savefig('visvalingam.png')
   print("saved visvalingam.png")
   #plot.show()
   



# -*- coding: utf-8 -*-
"""

@author: Pieter Roggemans

Based on:
    Python implementation of the Ramer-Douglas-Peucker algorithm.
    :copyright: 2014-2016 Fabian Hirschmann <fabian@hirschmann.email>
    :license: MIT, see LICENSE.txt for more details.
"""

from math import sqrt
from functools import partial
import numpy as np
import sys

if sys.version_info[0] >= 3:
    xrange = range


def pldist(point, start, end):
    """
    Calculates the distance from ``point`` to the line given
    by the points ``start`` and ``end``.
    :param point: a point
    :type point: numpy array
    :param start: a point of the line
    :type start: numpy array
    :param end: another point of the line
    :type end: numpy array
    """
    if np.all(np.equal(start, end)):
        return np.linalg.norm(point - start)

    return np.divide(
            np.abs(np.linalg.norm(np.cross(end - start, start - point))),
            np.linalg.norm(end - start))


def rdp_rec(M, epsilon, dist=pldist):
    """
    Simplifies a given array of points.
    Recursive version.
    :param M: an array
    :type M: numpy array
    :param epsilon: epsilon in the rdp algorithm
    :type epsilon: float
    :param dist: distance function
    :type dist: function with signature ``f(point, start, end)`` -- see :func:`rdp.pldist`
    """
    dmax = 0.0
    index = -1

    for i in xrange(1, M.shape[0]):
        d = dist(M[i], M[0], M[-1])

        if d > dmax:
            index = i
            dmax = d

    if dmax > epsilon:
        r1 = rdp_rec(M[:index + 1], epsilon, dist)
        r2 = rdp_rec(M[index:], epsilon, dist)

        return np.vstack((r1[:-1], r2))
    else:
        return np.vstack((M[0], M[-1]))


def _rdp_iter(M, start_index, last_index, epsilon, dist=pldist):
    stk = []
    stk.append([start_index, last_index])
    global_start_index = start_index
    indices = np.ones(last_index - start_index + 1, dtype=bool)

    while stk:
        start_index, last_index = stk.pop()

        dmax = 0.0
        index = start_index

        for i in xrange(index + 1, last_index):
            if indices[i - global_start_index]:
                d = dist(M[i], M[start_index], M[last_index])
                if d > dmax:
                    index = i
                    dmax = d

        if dmax > epsilon:
            stk.append([start_index, index])
            stk.append([index, last_index])
        else:
            for i in xrange(start_index + 1, last_index):
                indices[i - global_start_index] = False

    return indices


def rdp_iter(M, epsilon, dist=pldist, return_mask=False):
    """
    Simplifies a given array of points.
    Iterative version.
    :param M: an array
    :type M: numpy array
    :param epsilon: epsilon in the rdp algorithm
    :type epsilon: float
    :param dist: distance function
    :type dist: function with signature ``f(point, start, end)`` -- see :func:`rdp.pldist`
    :param return_mask: return the mask of points to keep instead
    :type return_mask: bool
    """
    mask = _rdp_iter(M, 0, len(M) - 1, epsilon, dist)

    if return_mask:
        return mask

    return M[mask]


def rdp(M, epsilon=0, dist=pldist, algo="iter", return_mask=False):
    """
    Simplifies a given array of points using the Ramer-Douglas-Peucker
    algorithm.
    Example:
    >>> from rdp import rdp
    >>> rdp([[1, 1], [2, 2], [3, 3], [4, 4]])
    [[1, 1], [4, 4]]
    This is a convenience wrapper around both :func:`rdp.rdp_iter` 
    and :func:`rdp.rdp_rec` that detects if the input is a numpy array
    in order to adapt the output accordingly. This means that
    when it is called using a Python list as argument, a Python
    list is returned, and in case of an invocation using a numpy
    array, a NumPy array is returned.
    The parameter ``return_mask=True`` can be used in conjunction
    with ``algo="iter"`` to return only the mask of points to keep. Example:
    >>> from rdp import rdp
    >>> import numpy as np
    >>> arr = np.array([1, 1, 2, 2, 3, 3, 4, 4]).reshape(4, 2)
    >>> arr
    array([[1, 1],
           [2, 2],
           [3, 3],
           [4, 4]])
    >>> mask = rdp(arr, algo="iter", return_mask=True)
    >>> mask
    array([ True, False, False,  True], dtype=bool)
    >>> arr[mask]
    array([[1, 1],
           [4, 4]])
    :param M: a series of points
    :type M: numpy array with shape ``(n,d)`` where ``n`` is the number of points and ``d`` their dimension
    :param epsilon: epsilon in the rdp algorithm
    :type epsilon: float
    :param dist: distance function
    :type dist: function with signature ``f(point, start, end)`` -- see :func:`rdp.pldist`
    :param algo: either ``iter`` for an iterative algorithm or ``rec`` for a recursive algorithm
    :type algo: string
    :param return_mask: return mask instead of simplified array
    :type return_mask: bool
    """

    if algo == "iter":
        algo = partial(rdp_iter, return_mask=return_mask)
    elif algo == "rec":
        if return_mask:
            raise NotImplementedError("return_mask=True not supported with algo=\"rec\"")
        algo = rdp_rec
        
    if "numpy" in str(type(M)):
        return algo(M, epsilon, dist)

    return algo(np.array(M), epsilon, dist).tolist()



# -*- coding: utf-8 -*-
"""
Module with specific helper functions to manage the logging of orthoseg.

TODO: maybe it is cleaner to replace most code here by a config dict?
"""

import logging
import datetime
from pathlib import Path

#-------------------------------------------------------------
# Init logging
#-------------------------------------------------------------

def main_log_init(log_dir: Path,
                  log_basefilename: str):

    # Check input parameters
    if not log_dir:
        raise Exception(f"Error: log_dir is mandatory!")
    
    # Make sure the log dir exists
    if not log_dir.exists():
        log_dir.mkdir(parents=True, exist_ok=True)
        
    # Get root logger
    logger = logging.getLogger('')
    
    # Set the general maximum log level...
    logger.setLevel(logging.INFO)
    for handler in logger.handlers:
        handler.flush()
        handler.close()
    
    # Remove all handlers and add the ones I want again, so a new log file is created for each run
    # Remark: the function removehandler doesn't seem to work?
    logger.handlers = []
    
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    #ch.setFormatter(logging.Formatter('%(levelname)s|%(name)s|%(message)s'))
    #ch.setFormatter(logging.Formatter('%(asctime)s|%(levelname)s|%(name)s|%(message)s',
    #                                  datefmt='%H:%M:%S,uuu'))
    ch.setFormatter(logging.Formatter(fmt='%(asctime)s.%(msecs)03d|%(levelname)s|%(name)s|%(message)s',
                                      datefmt='%H:%M:%S')) 
    logger.addHandler(ch)
    
    log_filepath = log_dir / f"{datetime.datetime.now():%Y-%m-%d_%H-%M-%S}_{log_basefilename}.log"
    fh = logging.FileHandler(filename=str(log_filepath))
    fh.setLevel(logging.INFO)
    fh.setFormatter(logging.Formatter('%(asctime)s|%(levelname)s|%(name)s|%(message)s'))
    logger.addHandler(fh)
    
    return logger

# If the script is ran directly...
if __name__ == '__main__':
    raise Exception("Not implemented")
    


# -*- coding: utf-8 -*-
"""
Module with generic usable utility functions to work with geo files.
"""

import filecmp
import logging
import os
from pathlib import Path
import shutil
from typing import List

# If environment not properly intitialised
#if 'PROJ_LIB' not in os.environ:
#os.environ['PROJ_LIB'] = r"C:\Tools\anaconda3\envs\test_rtree1\Library\share\proj"
#if 'GDAL_DATA' not in os.environ:
#os.environ['GDAL_DATA'] = r"C:\Tools\anaconda3\envs\test_rtree1\Library\share\gdal"
os.environ['PATH'] += r"C:\Tools\anaconda3\envs\test_rtree2\Library\bin" + os.pathsep + os.environ['GDAL_DATA']

import fiona
import geopandas as gpd

#-------------------------------------------------------------
# First define/init some general variables/constants
#-------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)
#logger.setLevel(logging.DEBUG)

#-------------------------------------------------------------
# The real work
#-------------------------------------------------------------

def read_file(filepath: Path,
              layer: str = None,
              columns: List[str] = None) -> gpd.GeoDataFrame:
    """
    Reads a file to a pandas dataframe. The fileformat is detected based on the filepath extension.

    # TODO: think about if possible/how to support  adding optional parameter and pass them to next function, example encoding, float_format,...
    """
    # Init
    ext_lower = filepath.suffix.lower()

    # For file multilayer types, if no layer name specified, check if there is only one layer in the file.
    if(ext_lower in ['.gpkg'] 
       and layer is None):
        listlayers = fiona.listlayers(filepath)
        if len(listlayers) == 1:
            layer = listlayers[0]
        else:
            raise Exception(f"File contains {len(listlayers)} layers: {listlayers}, but layer is not specified: {filepath}")

    # Depending on the extension... different implementations
    if ext_lower == '.shp':
        return gpd.read_file(str(filepath))
    elif ext_lower == '.geojson':
        return gpd.read_file(str(filepath))
    elif ext_lower == '.gpkg':
        return gpd.read_file(str(filepath), layer=layer)
    else:
        raise Exception(f"Not implemented for extension {ext_lower}")

def to_file(gdf: gpd.GeoDataFrame,
            filepath: Path,
            layer: str = None,
            index: bool = True):
    """
    Reads a pandas dataframe to file. The fileformat is detected based on the filepath extension.

    # TODO: think about if possible/how to support  adding optional parameter and pass them to next function, example encoding, float_format,...
    """
    # If no layer name specified, use the filename (without extension)
    if layer is None:
        layer = filepath.stem
    # If the dataframe is empty, log warning and return
    if len(gdf) <= 0:
        logger.warn(f"Cannot write an empty dataframe to {filepath}.{layer}")
        return

    # Depending on the extension... different implementations
    ext_lower = filepath.suffix.lower()
    if ext_lower == '.shp':
        gdf.to_file(str(filepath))
    elif ext_lower == '.geojson':
        gdf.to_file(str(filepath), driver='GeoJSON')
    elif ext_lower == '.gpkg':
        gdf.to_file(str(filepath), layer=layer, driver='GPKG')
    else:
        raise Exception(f"Not implemented for extension {ext_lower}")
        
def get_crs(filepath: Path):
    with fiona.open(str(filepath), 'r') as geofile:
        return geofile.crs

def cmp(filepath1: Path, filepath2: Path):
    
    # For a shapefile, multiple files need to be compared
    if filepath1.suffix.lower() == '.shp':
        filepath2_noext, _ = os.path.splitext(filepath2)
        shapefile_extentions = [".shp", ".dbf", ".shx"]
        filepath1_noext = filepath1.parent / filepath1.stem
        filepath2_noext = filepath2.parent / filepath2.stem
        for ext in shapefile_extentions:
            if not filecmp.cmp(str(filepath1_noext) + ext, str(filepath2_noext) + ext):
                logger.info(f"File {filepath1_noext}{ext} is differnet from {filepath2_noext}{ext}")
                return False
        return True
    else:
        return filecmp.cmp(str(filepath1), str(filepath2))
    
def copy(filepath_src: Path, dest: Path):

    # For a shapefile, multiple files need to be copied
    if filepath_src.suffix.lower() == '.shp':
        shapefile_extentions = [".shp", ".dbf", ".shx", ".prj"]

        # If dest is a dir, just use copy. Otherwise concat dest filepaths
        filepath_src_noext = filepath_src.parent / filepath_src.stem
        if dest.is_dir():
            for ext in shapefile_extentions:
                shutil.copy(str(filepath_src_noext) + ext, dest)
        else:
            filepath_dest_noext = dest.parent / dest.stem
            for ext in shapefile_extentions:
                shutil.copy(str(filepath_src_noext) + ext, str(filepath_dest_noext) + ext)                
    else:
        return shutil.copy(filepath_src, dest)

def get_driver(filepath: Path) -> str:
    """
    Get the driver to use for the file extension of this filepath.
    """
    return get_driver_for_ext(filepath.suffix)

def get_driver_for_ext(file_ext: str) -> str:
    """
    Get the driver to use for this file extension.
    """
    file_ext_lower = file_ext.lower()
    if file_ext_lower == '.shp':
        return 'ESRI Shapefile'
    elif file_ext_lower == '.geojson':
        return 'GeoJSON'
    elif file_ext_lower == '.gpkg':
        return 'GPKG'
    else:
        raise Exception(f"Not implemented for extension {file_ext_lower}")        



# -*- coding: utf-8 -*-
"""
Module with functions for post-processing prediction masks towards polygons.
"""

import argparse
from pathlib import Path
import shlex
import sys

from orthoseg.helpers import config_helper as conf
from orthoseg.helpers import log_helper
from orthoseg.lib import postprocess_predictions as postp

def postprocess_argstr(argstr):
    args = shlex.split(argstr)
    postprocess_args(args)

def postprocess_args(args):

    ##### Interprete arguments #####
    parser = argparse.ArgumentParser(add_help=False)

    # Required arguments
    required = parser.add_argument_group('Required arguments')
    required.add_argument("--configfile", type=str, required=True,
            help="The config file to use")
    
    # Optional arguments
    optional = parser.add_argument_group('Optional arguments')
    # Add back help 
    optional.add_argument('-h', '--help', action='help', default=argparse.SUPPRESS,
            help='Show this help message and exit')

    # Interprete arguments
    args = parser.parse_args(args)

    ##### Run! #####
    postprocess(projectconfig_path=Path(args.configfile))

def postprocess(
        projectconfig_path: Path,
        imagelayerconfig_path: Path = None):
    """
    Postprocess.

    Args:
        projectconfig_path (Path): Path to the projects config file.
        imagelayerconfig_path (Path, optional): Path to the imagelayer config file. If not specified, 
            the path specified in files.image_layers_config_filepath in the project config will be used. 
            Defaults to None.
    """

    ##### Init #####   
    # Load config
    config_filepaths = conf.search_projectconfig_files(projectconfig_path=projectconfig_path)
    conf.read_project_config(config_filepaths, imagelayerconfig_path)
    
    # Main initialisation of the logging
    global logger
    logger = log_helper.main_log_init(conf.dirs.getpath('log_training_dir'), __name__)      
    logger.debug(f"Config used: \n{conf.pformat_config()}")

    # Input dir = the "most recent" prediction result dir for this subject 
    prediction_basedir = Path(f"{conf.dirs['predict_image_output_basedir']}_{conf.general['segment_subject']}_")
    searchstring = f"{prediction_basedir.name}*/"
    prediction_dirs = sorted(prediction_basedir.parent.glob(searchstring), reverse=True)
    if len(prediction_dirs) == 0:
        raise Exception(f"STOP: No prediction dirs found with search string {searchstring} in {prediction_basedir.parent}")        
    input_dir = prediction_dirs[0]
	
    # Format output dir, partly based on input dir
    # Remove first 2 field from the input dir to get the model info
    input_dir_splitted = input_dir.name.split('_')
    model_info = []
    for i, input_dir_field in enumerate(input_dir_splitted):
        if i >= 2:
            model_info.append(input_dir_field)
    
    output_dir = conf.dirs.getpath('output_vector_dir') / conf.predict['image_layer']
    output_vector_name = f"{'_'.join(model_info)}_{conf.predict['image_layer']}"
    output_filepath = output_dir / f"{output_vector_name}.gpkg"
    
    ##### Go! #####
    border_pixels_to_ignore = conf.predict.getint('image_pixels_overlap')
    postp.postprocess_predictions(
            input_dir=input_dir,
            output_filepath=output_filepath,
            input_ext='.tif',
            border_pixels_to_ignore=border_pixels_to_ignore,
            evaluate_mode=False,
            force=False)

#-------------------------------------------------------------
# If the script is ran directly...
#-------------------------------------------------------------

if __name__ == '__main__':
    postprocess_args(sys.argv[1:])
    


# -*- coding: utf-8 -*-
"""
Unet definition in Keras.

Based on code from:
    * unet: https://github.com/zhixuhao/unet
    * preload first layers with VGG16 weights:
      https://www.microsoft.com/developerblog/2018/07/05/satellite-images-segmentation-sustainable-farming/

TODO: review code and cleanup due to introduction of model factory

"""

import logging
from tensorflow import keras as kr
#import keras as kr
import numpy as np

#-------------------------------------------------------------
# First define/init some general variables/constants
#-------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)

#-------------------------------------------------------------
# The real work
#-------------------------------------------------------------

def get_model(input_width=256, input_height=256, nb_channels=3, nb_classes=1,
              loss_mode='binary_crossentropy', learning_rate=0.0001,
              init_model_weights: bool = False, pretrained_weights_filepath: str = None):

    inputs = kr.layers.Input((input_width, input_height, nb_channels))
    conv1 = kr.layers.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_normal', name="block1_conv1")(inputs)
    conv1 = kr.layers.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_normal', name="block1_conv2")(conv1)
    pool1 = kr.layers.MaxPooling2D(pool_size=(2, 2))(conv1)
    conv2 = kr.layers.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_normal', name="block2_conv1")(pool1)
    conv2 = kr.layers.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_normal', name="block2_conv2")(conv2)
    pool2 = kr.layers.MaxPooling2D(pool_size=(2, 2))(conv2)
    conv3 = kr.layers.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_normal', name="block3_conv1")(pool2)
    conv3 = kr.layers.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_normal', name="block3_conv2")(conv3)
    pool3 = kr.layers.MaxPooling2D(pool_size=(2, 2))(conv3)
    conv4 = kr.layers.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_normal', name="block4_conv1")(pool3)
    conv4 = kr.layers.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_normal', name="block4_conv2")(conv4)
    drop4 = kr.layers.Dropout(0.5)(conv4)
    pool4 = kr.layers.MaxPooling2D(pool_size=(2, 2))(drop4)

    conv5 = kr.layers.Conv2D(1024, 3, activation='relu', padding='same', kernel_initializer='he_normal')(pool4)
    conv5 = kr.layers.Conv2D(1024, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv5)
    drop5 = kr.layers.Dropout(0.5)(conv5)

    up6 = kr.layers.Conv2D(512, 2, activation='relu', padding='same', kernel_initializer='he_normal')(kr.layers.UpSampling2D(size = (2,2))(drop5))
    merge6 = kr.layers.Concatenate(axis=3)([drop4, up6])
    conv6 = kr.layers.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_normal')(merge6)
    conv6 = kr.layers.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv6)

    up7 = kr.layers.Conv2D(256, 2, activation='relu', padding='same', kernel_initializer='he_normal')(kr.layers.UpSampling2D(size = (2,2))(conv6))
    merge7 = kr.layers.Concatenate(axis=3)([conv3, up7])
    conv7 = kr.layers.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_normal')(merge7)
    conv7 = kr.layers.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv7)

    up8 = kr.layers.Conv2D(128, 2, activation='relu', padding='same', kernel_initializer='he_normal')(kr.layers.UpSampling2D(size = (2,2))(conv7))
    merge8 = kr.layers.Concatenate(axis=3)([conv2, up8])
    conv8 = kr.layers.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_normal')(merge8)
    conv8 = kr.layers.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv8)

    up9 = kr.layers.Conv2D(64, 2, activation='relu', padding='same', kernel_initializer='he_normal')(kr.layers.UpSampling2D(size = (2,2))(conv8))
    merge9 = kr.layers.Concatenate(axis=3)([conv1, up9])
    conv9 = kr.layers.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_normal')(merge9)
    conv9 = kr.layers.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv9)
    conv9 = kr.layers.Conv2D(2, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv9)
    conv10 = kr.layers.Conv2D(nb_classes, 1, activation='sigmoid')(conv9)

    model = kr.models.Model(inputs=inputs, outputs=conv10)

    if loss_mode == "bcedice":
        loss_func = dice_coef_loss_bce
    elif loss_mode == "binary_crossentropy":
        loss_func = 'binary_crossentropy'
    else:
        raise Exception(f"Unknown loss function: {loss_mode}")

    # Default learning rate for Adam: lr=1e-3, but doesn't seem to work well for unet
    # TODO: probably this needs to be removed because of the introduction of 
    # the model factory
    model.compile(optimizer=kr.optimizers.Adam(lr=learning_rate), loss=loss_func,
                  metrics=[jaccard_coef, jacard_coef_flat,
                           jaccard_coef_int, dice_coef, 'accuracy'])

    if init_model_weights:
        # Load the VGG16 model with pretrained weights
        # Remark: this only has 3 channels. If the input images have more channels, those channels
        #         won't be pre-loaded
        vgg = kr.applications.vgg16.VGG16(include_top=False, input_shape=(input_width, input_height, 3), weights='imagenet')

        '''
        # Create empty array with the dimensions of the first convolution layer
        conv1_weights = np.zeros((3, 3, nb_channels, 64), dtype="float32")
        conv1_weights[:, :, :3, :] = vgg.get_layer("block1_conv1").get_weights()[0][:, :, :, :]
        bias = vgg.get_layer("block1_conv1").get_weights()[1]
        model.get_layer('block1_conv1').set_weights((conv1_weights, bias))
        '''

        #logger.info(f"VGG16 summary:")
        #vgg.summary()

        # Copy the weigths first to a new array to have flexibility in nb
        # channels of unet compared to the vgg16
        block1_conv1_weights = np.zeros((3, 3, nb_channels, 64), dtype="float32")
        block1_conv1_weights[:, :, :3, :] = vgg.get_layer("block1_conv1").get_weights()[0][:, :, :, :]
        bias = vgg.get_layer("block1_conv1").get_weights()[1]
        model.get_layer('block1_conv1').set_weights((block1_conv1_weights, bias))

        block1_conv2_weights = np.zeros((3, 3, 64, 64), dtype="float32")
        block1_conv2_weights[:, :, :, :] = vgg.get_layer("block1_conv2").get_weights()[0][:, :, :, :]
        bias = vgg.get_layer("block1_conv2").get_weights()[1]
        model.get_layer('block1_conv2').set_weights((block1_conv2_weights, bias))

        block2_conv1_weights = np.zeros((3, 3, 64, 128), dtype="float32")
        block2_conv1_weights[:, :, :, :] = vgg.get_layer("block2_conv1").get_weights()[0][:, :, :, :]
        bias = vgg.get_layer("block2_conv1").get_weights()[1]
        model.get_layer('block2_conv1').set_weights((block2_conv1_weights, bias))

        block2_conv2_weights = np.zeros((3, 3, 128, 128), dtype="float32")
        block2_conv2_weights[:, :, :, :] = vgg.get_layer("block2_conv2").get_weights()[0][:, :, :, :]
        bias = vgg.get_layer("block2_conv2").get_weights()[1]
        model.get_layer('block2_conv2').set_weights((block2_conv2_weights, bias))

    #model.summary()

    if pretrained_weights_filepath:
        '''
        logger.info(f"Load model from {pretrained_weights_filepath}")
        
        model = kr.models.load_model(
                pretrained_weights_filepath,
                custom_objects={'jaccard_coef': jaccard_coef,
                                'jacard_coef_flat': jacard_coef_flat,
                                'jaccard_coef_int': jaccard_coef_int,
                                'dice_coef': dice_coef,
                                'dice_coef_loss_bce': dice_coef_loss_bce})
        
        model = get_unet(input_width=input_width, input_height=input_height,
                                 nb_channels=nb_channels, nb_classes=nb_classes, loss_mode=loss_mode,
                                 init_with_vgg16=False, pretrained_weights_filepath=None)

        model_preload = get_unet(input_width=input_width, input_height=input_height,
                                 nb_channels=nb_channels, nb_classes=nb_classes, loss_mode=loss_mode,
                                 init_with_vgg16=False, pretrained_weights_filepath=None)
        logger.info(f"Preload weights from {pretrained_weights_filepath}")
        model_preload.load_weights(pretrained_weights_filepath)
        
        for layer_preloaded, layer_built in zip(model_preload.layers, model.layers):
            layer_built.set_weights(layer_preloaded.get_weights())
        '''
        
        logger.info(f"Preload weights from {pretrained_weights_filepath}")
        model.load_weights(pretrained_weights_filepath)
        
    return model

def load_unet_model(filepath: str,
                    learning_rate: float = None):
    logger.info(f"Load model from {filepath}")
    model = kr.models.load_model(filepath,
                                 custom_objects={'jaccard_coef': jaccard_coef,
                                                 'jacard_coef_flat': jacard_coef_flat,
                                                 'jaccard_coef_int': jaccard_coef_int,
                                                 'dice_coef': dice_coef,
                                                 'dice_coef_loss_bce': dice_coef_loss_bce})
    if learning_rate:
        kr.backend.set_value(model.optimizer.lr, learning_rate)
        
    return model

#------------------------------------------
# Loss functions
#------------------------------------------

def dice_coef_loss(y_true, y_pred):
    return 1 - dice_coef(y_true, y_pred)

def bootstrapped_crossentropy(y_true, y_pred, bootstrap_type='hard', alpha=0.95):
    target_tensor = y_true
    prediction_tensor = y_pred
    _epsilon = kr.backend.tensorflow_backend._to_tensor(kr.backend.epsilon(), prediction_tensor.dtype.base_dtype)
    prediction_tensor = kr.backend.tf.clip_by_value(prediction_tensor, _epsilon, 1 - _epsilon)
    prediction_tensor = kr.backend.tf.log(prediction_tensor / (1 - prediction_tensor))

    if bootstrap_type == 'soft':
        bootstrap_target_tensor = alpha * target_tensor + (1.0 - alpha) * kr.backend.tf.sigmoid(prediction_tensor)
    else:
        bootstrap_target_tensor = alpha * target_tensor + (1.0 - alpha) * kr.backend.tf.cast(
            kr.backend.tf.sigmoid(prediction_tensor) > 0.5, kr.backend.tf.float32)
    return kr.backend.mean(kr.backend.tf.nn.sigmoid_cross_entropy_with_logits(
        labels=bootstrap_target_tensor, logits=prediction_tensor))

def dice_coef_loss_bce(y_true, y_pred):
    dice = 0.5
    bce = 0.5
    bootstrapping = 'hard'
    alpha = 1.
    return bootstrapped_crossentropy(y_true, y_pred, bootstrapping, alpha) * bce + dice_coef_loss(y_true, y_pred) * dice

#------------------------------------------
# Metrics functions
#------------------------------------------

SMOOTH_LOSS = 1e-12

def jaccard_coef(y_true, y_pred):
    intersection = kr.backend.sum(y_true * y_pred, axis=[0, -1, -2])
    sum_ = kr.backend.sum(y_true + y_pred, axis=[0, -1, -2])

    jac = (intersection + SMOOTH_LOSS) / (sum_ - intersection + SMOOTH_LOSS)

    return kr.backend.mean(jac)

def jaccard_coef_int(y_true, y_pred):
    y_pred_pos = kr.backend.round(kr.backend.clip(y_pred, 0, 1))

    intersection = kr.backend.sum(y_true * y_pred_pos, axis=[0, -1, -2])
    sum_ = kr.backend.sum(y_true + y_pred_pos, axis=[0, -1, -2])
    jac = (intersection + SMOOTH_LOSS) / (sum_ - intersection + SMOOTH_LOSS)
    return kr.backend.mean(jac)

def jacard_coef_flat(y_true, y_pred):
    y_true_f = kr.backend.flatten(y_true)
    y_pred_f = kr.backend.flatten(y_pred)
    intersection = kr.backend.sum(y_true_f * y_pred_f)
    return (intersection + SMOOTH_LOSS) / (kr.backend.sum(y_true_f) + kr.backend.sum(y_pred_f) - intersection + SMOOTH_LOSS)

def dice_coef(y_true, y_pred, smooth=1.0):
    y_true_f = kr.backend.flatten(y_true)
    y_pred_f = kr.backend.flatten(y_pred)
    intersection = kr.backend.sum(y_true_f * y_pred_f)
    return (2. * intersection + smooth) / (kr.backend.sum(y_true_f) + kr.backend.sum(y_pred_f) + smooth)

def pct_wrong(y_true, y_pred):
    y_pred_pos = kr.backend.round(kr.backend.clip(y_pred, 0, 1))

    intersection = kr.backend.sum(y_true * y_pred_pos, axis=[0, -1, -2])
    sum_ = kr.backend.sum(y_true + y_pred_pos, axis=[0, -1, -2])
    jac = (intersection + SMOOTH_LOSS) / (sum_ - intersection + SMOOTH_LOSS)
    return kr.backend.mean(jac)



# -*- coding: utf-8 -*-
"""
High-level API to run a segmentation.
"""

import argparse
from pathlib import Path
import shlex
import sys

#import os
#os.environ["CUDA_VISIBLE_DEVICES"] = "-1" # Disable using GPU
import tensorflow as tf
from tensorflow import keras as kr  

from orthoseg.helpers import config_helper as conf
from orthoseg.helpers import log_helper
from orthoseg.lib import predicter
import orthoseg.model.model_factory as mf
import orthoseg.model.model_helper as mh

def predict_argstr(argstr):
    args = shlex.split(argstr)
    predict_args(args)

def predict_args(args):

    ##### Interprete arguments #####
    parser = argparse.ArgumentParser(add_help=False)

    # Required arguments
    required = parser.add_argument_group('Required arguments')
    required.add_argument("--configfile", type=str, required=True,
            help="The config file to use")
    
    # Optional arguments
    optional = parser.add_argument_group('Optional arguments')
    # Add back help 
    optional.add_argument('-h', '--help', action='help', default=argparse.SUPPRESS,
            help='Show this help message and exit')

    # Interprete arguments
    args = parser.parse_args(args)

    ##### Run! #####
    predict(projectconfig_path=Path(args.configfile))

def predict(
        projectconfig_path: Path,
        imagelayerconfig_path: Path = None):
    """
    Run a prediction of the input dir given.

    Args:
        projectconfig_path (Path): Path to the projects config file.
        imagelayerconfig_path (Path, optional): Path to the imagelayer config file. If not specified, 
            the path specified in files.image_layers_config_filepath in the project config will be used. 
            Defaults to None.
    """
    ##### Init #####   
    # Load config
    ##### Init #####   
    # Load config
    config_filepaths = conf.search_projectconfig_files(projectconfig_path=projectconfig_path)
    conf.read_project_config(config_filepaths, imagelayerconfig_path)
    
    # Main initialisation of the logging
    global logger
    logger = log_helper.main_log_init(conf.dirs.getpath('log_training_dir'), __name__)      
    logger.debug(f"Config used: \n{conf.pformat_config()}")
    
    # Check if the input inmages dir exists
    input_image_dir = conf.dirs.getpath('predict_image_input_dir')
    if not input_image_dir.exists():
        raise Exception(f"STOP: input image dir doesn't exist: {input_image_dir}")
    
    # TODO: add something to delete old data, predictions???

    # Create base filename of model to use
    # TODO: is force data version the most logical, or rather implement 
    #       force weights file or ?
    force_model_traindata_id = conf.train.getint('force_model_traindata_id')
    if force_model_traindata_id > -1:
        traindata_id = force_model_traindata_id 
    else:
        traindata_id = mh.get_max_data_version(model_dir=conf.dirs.getpath('model_dir'))
        #logger.info(f"max model_traindata_id found: {model_traindata_id}")
    
    # Get the best model that already exists for this train dataset
    trainparams_id = conf.train.getint('trainparams_id')
    best_model = mh.get_best_model(
            model_dir=conf.dirs.getpath('model_dir'), 
            segment_subject=conf.general['segment_subject'],
            traindata_id=traindata_id,
            trainparams_id=trainparams_id)
    
    # Check if a model was found
    if best_model is None:
        message = f"No model found in model_dir: {conf.dirs.getpath('model_dir')} for traindata_id: {traindata_id}"
        logger.critical(message)
        raise Exception(message)
    else:    
        model_weights_filepath = best_model['filepath']
        logger.info(f"Best model found: {model_weights_filepath}")
    
    # Prepare output subdir to be used for predictions
    predict_out_subdir = f"{best_model['basefilename']}"
    if trainparams_id > 0:
        predict_out_subdir += f"_{trainparams_id}"
    predict_out_subdir += f"_{best_model['epoch']}"
    
    # Try optimizing model with tensorrt
    try:
        # Try import
        from tensorflow.python.compiler.tensorrt import trt_convert as trt

        # Import didn't fail, so optimize model
        logger.info('Tensorrt is available, so use optimized model')
        savedmodel_optim_dir = best_model['filepath'].parent / best_model['filepath'].stem + "_optim"
        if not savedmodel_optim_dir.exists():
            # If base model not yet in savedmodel format
            savedmodel_dir = best_model['filepath'].parent / best_model['filepath'].stem
            if not savedmodel_dir.exists():
                logger.info(f"SavedModel format not yet available, so load model + weights from {best_model['filepath']}")
                model = mf.load_model(best_model['filepath'], compile=False)
                logger.info(f"Now save again as savedmodel to {savedmodel_dir}")
                tf.saved_model.save(model, str(savedmodel_dir))
                del model

            # Now optimize model
            logger.info(f"Optimize + save model to {savedmodel_optim_dir}")
            converter = trt.TrtGraphConverterV2(
                    input_saved_model_dir=savedmodel_dir,
                    is_dynamic_op=True,
                    precision_mode='FP16')
            converter.convert()
            converter.save(savedmodel_optim_dir)
        
        logger.info(f"Load optimized model + weights from {savedmodel_optim_dir}")
        model = tf.keras.models.load_model(savedmodel_optim_dir)

    except ImportError as e:
        logger.info('Tensorrt is not available, so load unoptimized model')
        model = mf.load_model(best_model['filepath'], compile=False)

    # Prepare the model for predicting
    nb_gpu = len(tf.config.experimental.list_physical_devices('GPU'))
    batch_size = conf.predict.getint('batch_size')
    # TODO: because of bug in tensorflow 1.14, multi GPU doesn't work (this way), 
    # so always use one
    if nb_gpu <= 1:
        model_for_predict = model
        logger.info(f"Predict using single GPU or CPU, with nb_gpu: {nb_gpu}")
    else:
        # If multiple GPU's available, create multi_gpu_model
        try:
            model_for_predict = kr.utils.multi_gpu_model(model, gpus=nb_gpu, cpu_relocation=True)
            logger.info(f"Predict using multiple GPUs: {nb_gpu}, batch size becomes: {batch_size*nb_gpu}")
            batch_size *= nb_gpu
        except ValueError:
            logger.info("Predict using single GPU or CPU")
            model_for_predict = model

    # Predict for entire dataset
    image_layer = conf.image_layers[conf.predict['image_layer']]
    predict_output_dir = Path(f"{conf.dirs['predict_image_output_basedir']}_{predict_out_subdir}")
    predicter.predict_dir(
            model=model_for_predict,
            input_image_dir=input_image_dir,
            output_base_dir=predict_output_dir,
            border_pixels_to_ignore=conf.predict.getint('image_pixels_overlap'),
            projection_if_missing=image_layer['projection'],
            input_mask_dir=None,
            batch_size=batch_size,
            evaluate_mode=False)

# If the script is ran directly...
if __name__ == '__main__':
    predict_args(sys.argv[1:])
    


# -*- coding: utf-8 -*-
"""
Module with generic usable utility functions to make some tasks using OWS services easier.
"""

import concurrent.futures as futures
import datetime
import itertools as it
import logging
import math
import os
from pathlib import Path
import random
import time
from typing import List, Optional, Tuple

import owslib
import owslib.wms
import rasterio as rio
import geopandas as gpd
import shapely.geometry as sh_geom

from orthoseg.util import vector_util

#-------------------------------------------------------------
# First define/init some general variables/constants
#-------------------------------------------------------------
FORMAT_GEOTIFF = 'image/geotiff'
FORMAT_GEOTIFF_EXT = '.tif'
FORMAT_GEOTIFF_EXT_WORLD = '.tfw'

FORMAT_TIFF = 'image/tiff'
FORMAT_TIFF_EXT = '.tif'
FORMAT_TIFF_EXT_WORLD = '.tfw'

FORMAT_JPEG = 'image/jpeg'
FORMAT_JPEG_EXT = '.jpg'
FORMAT_JPEG_EXT_WORLD = '.jgw'

FORMAT_PNG = 'image/png'
FORMAT_PNG_EXT = '.png'
FORMAT_PNG_EXT_WORLD = '.pgw'

# Get a logger...
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

#-------------------------------------------------------------
# The real work
#-------------------------------------------------------------

def get_images_for_grid(
        wms_server_url: str,
        wms_version: str,
        wms_layernames: List[str],
        srs: str,
        output_image_dir: Path,
        image_gen_bounds: Tuple[float, float, float, float] = None,
        image_gen_roi_filepath: Optional[Path] = None,
        grid_xmin: float = 0.0,
        grid_ymin: float = 0.0,
        image_srs_pixel_x_size: float = 0.25,
        image_srs_pixel_y_size: float = 0.25,
        image_pixel_width: int = 1024,
        image_pixel_height: int = 1024,
        image_pixels_ignore_border: int = 0,
        nb_concurrent_calls: int = 1,
        random_sleep: float = 0.0,        
        image_format: str = FORMAT_GEOTIFF,
        image_format_save: str = None,
        tiff_compress: str = 'lzw',
        transparent: bool = False,
        wms_layerstyles: List[str] = ['default'],
        pixels_overlap: int = 0,
        column_start: int = 0,
        nb_images_to_skip: int = None,
        max_nb_images: int = -1,
        force: bool = False):
    
    ##### Init #####
    if image_format_save is None:
        image_format_save = image_format

    srs_width = math.fabs(image_pixel_width*image_srs_pixel_x_size)   # tile width in units of crs => 500 m
    srs_height = math.fabs(image_pixel_height*image_srs_pixel_y_size) # tile height in units of crs => 500 m
    is_srs_projected = rio.crs.CRS.from_string(srs).is_projected
    
    # Read the region of interest file if provided
    roi_gdf = None
    if image_gen_roi_filepath is not None:
        # Open vector layer
        logger.info(f"Open vector file {image_gen_roi_filepath}")
        roi_gdf = gpd.read_file(str(image_gen_roi_filepath))
                
        # If the generate_window wasn't specified, calculate the bounds
        # based on the roi (but make sure they fit the grid!)
        if not image_gen_bounds:
            roi_bounds = roi_gdf.geometry.total_bounds
            image_gen_bounds = (roi_bounds[0]-((roi_bounds[0]-grid_xmin)%srs_width),
                                roi_bounds[1]-((roi_bounds[1]-grid_ymin)%srs_height),
                                roi_bounds[2]+(grid_xmin-((roi_bounds[2]-grid_xmin)%srs_width)),
                                roi_bounds[3]+(grid_ymin-((roi_bounds[3]-grid_ymin)%srs_height)))
            logger.info(f"roi_bounds: {roi_bounds}, image_gen_bounds: {image_gen_bounds}")
        
        # If there are large objects in the roi, segment them to speed up
        # TODO: implement usefull check to see if segmenting is usefull...
        # TODO: support creating a grid in latlon????
        if is_srs_projected:
            
            # Create grid
            grid = vector_util.create_grid(
                    xmin=image_gen_bounds[0], ymin=image_gen_bounds[1], 
                    xmax=image_gen_bounds[2], ymax=image_gen_bounds[3],
                    cell_width=4000, cell_height=4000)
            # Create intersection layer between grid and roi
            #roi_gdf = roi_gdf.geometry.intersection(grid.geometry)
            roi_gdf = gpd.overlay(grid, roi_gdf, how='intersection')
            
            # Explode possible multipolygons to polygons
            roi_gdf = roi_gdf.reset_index(drop=True).explode()
                        
            #roi_gdf.to_file("X:\\Monitoring\\OrthoSeg\\roi_gridded.shp")
            
    # Check if the image_gen_bounds are compatible with the grid...
    if (image_gen_bounds[0]-grid_xmin)%srs_width != 0:
        xmin_new = image_gen_bounds[0] - ((image_gen_bounds[0]-grid_xmin)%srs_width)
        logger.warning(f"xmin {image_gen_bounds[0]} in compatible with grid, {xmin_new} will be used")
        image_gen_bounds = (xmin_new, image_gen_bounds[1], image_gen_bounds[2], image_gen_bounds[3])
    if (image_gen_bounds[1]-grid_ymin)%srs_height != 0:
        ymin_new = image_gen_bounds[1] - ((image_gen_bounds[1]-grid_ymin)%srs_height)
        logger.warning(f"ymin {image_gen_bounds[1]} incompatible with grid, {ymin_new} will be used")
        image_gen_bounds = (image_gen_bounds[0], ymin_new, image_gen_bounds[2], image_gen_bounds[3])
    if (image_gen_bounds[2]-grid_xmin)%srs_width != 0:
        xmax_new = image_gen_bounds[2] + srs_width - ((image_gen_bounds[2]-grid_xmin)%srs_width)
        logger.warning(f"xmax {image_gen_bounds[2]} incompatible with grid, {xmax_new} will be used")
        image_gen_bounds = (image_gen_bounds[0], image_gen_bounds[1], xmax_new, image_gen_bounds[3])
    if (image_gen_bounds[3]-grid_ymin)%srs_height != 0:
        ymax_new = image_gen_bounds[3] + srs_height - ((image_gen_bounds[3]-grid_ymin)%srs_height)
        logger.warning(f"ymax {image_gen_bounds[3]} incompatible with grid, {ymax_new} will be used")
        image_gen_bounds = (image_gen_bounds[0], image_gen_bounds[1], image_gen_bounds[2], ymax_new)

    # Calculate width and height...
    dx = math.fabs(image_gen_bounds[0] - image_gen_bounds[2]) # area width in units of crs
    dy = math.fabs(image_gen_bounds[1] - image_gen_bounds[3]) # area height in units of crs
    cols = int(math.ceil(dx / srs_width)) + 1
    rows = int(math.ceil(dy / srs_height)) + 1

    # Inits to start getting images 
    if not output_image_dir.exists():
        output_image_dir.mkdir(parents=True)
    auth = owslib.util.Authentication()
    wms = owslib.wms.WebMapService(wms_server_url, version=wms_version, auth=auth)

    with futures.ThreadPoolExecutor(nb_concurrent_calls) as pool:

        # Loop through all columns and get the images...
        start_time = None
        nb_todo = cols*rows
        nb_processed = 0
        nb_downloaded = 0
        nb_ignore_in_progress = 0
        bbox_list = []
        size_list = []
        output_filename_list = []
        for col in range(column_start, cols):
            
            image_xmin = col * srs_width + image_gen_bounds[0]
            image_xmax = (col + 1) * srs_width + image_gen_bounds[0]
    
            # If overlapping images are wanted... increase image bbox
            if pixels_overlap:
                image_xmin = image_xmin-(pixels_overlap*image_srs_pixel_x_size)
                image_xmax = image_xmax+(pixels_overlap*image_srs_pixel_x_size)
                
            # Put all the images of this column in a dir
            if is_srs_projected:
                output_dir = output_image_dir / f"{image_xmin:06.0f}"
            else:
                output_dir = output_image_dir / f"{image_xmin:09.4f}"
            if not output_dir.exists():
                output_dir.mkdir()
                
            logger.info(f"Start processing column {col}")
            for row in range(0, rows):
                nb_processed += 1
                
                # To be able to quickly get images spread over the roi...
                if(nb_images_to_skip
                   and (nb_processed%nb_images_to_skip) != 0):
                    #logger.debug(f"Skip this image, because {nb_processed}%{nb_images_to_skip} is not 0")
                    continue
    
                # Calculate y bounds
                image_ymin = row * srs_height + image_gen_bounds[1]
                image_ymax = (row + 1) * srs_height + image_gen_bounds[1]
    
                # If overlapping images are wanted... increase image bbox
                if pixels_overlap:
                    image_ymin = image_ymin-(pixels_overlap*image_srs_pixel_y_size)
                    image_ymax = image_ymax+(pixels_overlap*image_srs_pixel_y_size)
    
                # Create output filename
                output_filename = create_filename(
                        srs=srs,
                        bbox=(image_xmin, image_ymin, image_xmax, image_ymax),
                        size=(image_pixel_width+2*pixels_overlap, 
                              image_pixel_height+2*pixels_overlap),
                        image_format=image_format,
                        layername=None)
                output_filepath = output_dir / output_filename
                if not force and output_filepath.exists():
                    nb_ignore_in_progress += 1
                    logger.debug("    -> image exists already, so skip")
                    continue
                
                # If roi was provided, check first if the current image overlaps
                # TODO: using an chopped up version of the ROI is probably faster
                # TODO: possibly checking if the dest file exists already before 
                # this is faster...
                
                if roi_gdf is not None:
                    image_shape = sh_geom.box(image_xmin, image_ymin, image_xmax, image_ymax)
                    
                    #intersections = gpd.sjoin(roi_gdf, lines, how="inner", op='intersects')
                    spatial_index = roi_gdf.sindex
                    possible_matches_index = list(spatial_index.intersection(image_shape.bounds))
                    possible_matches = roi_gdf.iloc[possible_matches_index]
                    precise_matches = possible_matches[possible_matches.intersects(image_shape)]
                    
                    if len(precise_matches) == 0:
                        nb_ignore_in_progress += 1
                        logger.debug("    -> image doesn't overlap with roi, so skip")
                        continue                                       
                    
                    # TODO: cleanup old code
                    '''
                    if not roi_gdf.intersects(image_shape).any():
                        nb_ignore_in_progress += 1
                        logger.debug("    -> image doesn't overlap with roi, so skip")
                        continue                                       
                    '''
                    '''
                    intersects = False
                    # Loop through the roi rows. If one of them overlaps -> OK.
                    for roi_row in roi_rows:
                        # Check first on bounds to improve performance
                        if(roi_row['geom_bounds'].intersection(image_shape)
                           and roi_row['geom'].intersection(image_shape)):
                            intersects = True
                            break
                    if not intersects:
                        nb_ignore_in_progress += 1
                        logger.debug("    -> image doesn't overlap with roi, so skip")
                        continue
                    '''
                
                # Now we are getting to start fetching images... init start_time 
                if start_time is None:
                    start_time = datetime.datetime.now()
                
                # Append the image info to the batch arrays so they can be treated in 
                # bulk if the batch size is reached
                bbox_list.append((image_xmin, image_ymin, image_xmax, image_ymax))
                size_list.append((image_pixel_width+2*pixels_overlap, 
                                  image_pixel_height+2*pixels_overlap))
                output_filename_list.append(output_filename)
                
                # If the batch size is reached or we are at the last images
                nb_images_in_batch = len(bbox_list)
                if(nb_images_in_batch == nb_concurrent_calls or nb_processed == (nb_todo-1)):
    
                    start_time_batch_read = datetime.datetime.now()
                    
                    # Exec in parallel 
                    read_results = pool.map(
                            getmap_to_file,        # Function 
                            it.repeat(wms),
                            it.repeat(wms_layernames),
                            it.repeat(output_dir),
                            it.repeat(srs),
                            bbox_list,
                            size_list,
                            it.repeat(image_format),
                            it.repeat(image_format_save),
                            output_filename_list,
                            it.repeat(transparent),
                            it.repeat(tiff_compress),
                            it.repeat(wms_layerstyles),
                            it.repeat(random_sleep),
                            it.repeat(image_pixels_ignore_border),
                            it.repeat(force))
                            
                    for _ in read_results:
                        nb_downloaded += 1
                    
                    # Progress
                    logger.debug(f"Process image {nb_processed} out of {cols*rows}: {nb_processed/(cols*rows):.2f} %")

                    # Log the progress and prediction speed
                    time_passed_s = (datetime.datetime.now()-start_time).total_seconds()
                    time_passed_lastbatch_s = (datetime.datetime.now()-start_time_batch_read).total_seconds()
                    if time_passed_s > 0 and time_passed_lastbatch_s > 0:
                        nb_per_hour = ((nb_processed-nb_ignore_in_progress)/time_passed_s) * 3600
                        nb_per_hour_lastbatch = (nb_images_in_batch/time_passed_lastbatch_s) * 3600
                        hours_to_go = (int)((nb_todo - nb_processed)/nb_per_hour)
                        min_to_go = (int)((((nb_todo - nb_processed)/nb_per_hour)%1)*60)
                        print(f"\r{hours_to_go:3d}:{min_to_go:2d} left for {nb_todo-nb_processed} images at {nb_per_hour:0.0f}/h ({nb_per_hour_lastbatch:0.0f}/h last batch), with {nb_ignore_in_progress} skipped",
                              end='', flush=True)
                    
                    # Reset variable for next batch
                    bbox_list = []
                    size_list = []
                    output_filename_list = []

                    if max_nb_images > -1 and nb_downloaded >= max_nb_images:
                        return            
                    '''
                    # TODO: cleanup
                    #logger.info(f"Nb processed: {nb_processed} of {cols*rows} ({nb_processed/(cols*rows):.2f}%), Nb downloaded: {counter_downloaded}")
                    # Log the progress and prediction speed once 10 seconds have passed
                    secs_since_last_log = (datetime.datetime.now()-last_progress_log_time).total_seconds()
                    if secs_since_last_log > 10:
                        last_progress_log_time = datetime.datetime.now()
                        secs_passed_total = (datetime.datetime.now()-start_time).total_seconds()
                        nb_per_hour = ((nb_processed-nb_ignore_in_progress)/secs_passed_total) * 3600
                        hours_to_go = (int)((nb_todo-nb_processed)/nb_per_hour)
                        min_to_go = (int)((((nb_todo-nb_processed)/nb_per_hour)%1)*60)
                        print(f"\r{hours_to_go}:{min_to_go} left for {nb_todo-nb_processed} todo at {nb_per_hour:0.0f}/h",
                              end="", flush=True)
                    '''
                    
def getmap_to_file(
        wms: owslib.wms.WebMapService,
        layers: List[str],
        output_dir: Path,
        srs: str,
        bbox,
        size,
        image_format: str = FORMAT_GEOTIFF,
        image_format_save: str = None,
        output_filename: str = None,
        transparent: bool = False,
        tiff_compress: str = 'lzw',
        styles: List[str] = ['default'],
        random_sleep: float = 0.0,
        image_pixels_ignore_border: int = 0,
        force: bool = False,
        layername_in_filename: bool = False) -> Optional[Path]:
    """

    Args
        random_sleep: sleep a random time between 0 and this amount of seconds
                      between requests tot the WMS server
    """
    ##### Init #####
    # If no seperate save format is specified, use the standard image_format
    if image_format_save is None:
        image_format_save = image_format

    # If there isn't a filename supplied, create one...
    if output_filename is None:
        layername = None
        if layername_in_filename:
            layername = '_'.join(layers)
        output_filename = create_filename(
                srs=srs, 
                bbox=bbox, size=size, 
                image_format=image_format_save,
                layername=layername)

    # Create full output filepath
    output_filepath = output_dir / output_filename

    # If force is false and file exists already, stop...
    if force == False and output_filepath.exists():
        logger.debug(f"File already exists, skip: {output_filepath}")
        return None

    logger.debug(f"Get image to {output_filepath}")

    ##### Get image #####
    # Retry 10 times...
    nb_retries = 0
    time_sleep = 0
    while True:
        try:
            logger.debug(f"Start call GetMap for bbox {bbox}")

            # Some hacks for special cases...
            bbox_for_getmap = bbox
            size_for_getmap = size
            # Dirty hack to ask a bigger picture, and then remove the border again!
            if image_pixels_ignore_border > 0:
                x_pixsize = (bbox[2]-bbox[0])/size[0]
                y_pixsize = (bbox[3]-bbox[1])/size[1]
                bbox_for_getmap = (bbox[0] - x_pixsize*image_pixels_ignore_border,
                                   bbox[1] - y_pixsize*image_pixels_ignore_border,
                                   bbox[2] + x_pixsize*image_pixels_ignore_border,
                                   bbox[3] + y_pixsize*image_pixels_ignore_border)
                size_for_getmap = (size[0] + 2*image_pixels_ignore_border,
                                   size[1] + 2*image_pixels_ignore_border)
            # Dirty hack to support y,x cordinate system
            if srs.lower() == 'epsg:3059':
                bbox_for_getmap = (bbox_for_getmap[1], bbox_for_getmap[0], 
                                   bbox_for_getmap[3], bbox_for_getmap[2])

            response = wms.getmap(
                    layers=layers,
                    styles=styles,
                    srs=srs,
                    bbox=bbox_for_getmap,
                    size=size_for_getmap,
                    format=image_format,
                    transparent=transparent)
            logger.debug(f"Finished doing request {response.geturl()}")
            
            # If a random sleep was specified... apply it
            if random_sleep:
                time.sleep(random.uniform(0, random_sleep))
            
            # Image was retrieved... so stop loop
            break

        except Exception as ex:
            # Retry 10 times... and increase sleep time every time
            if nb_retries < 10:
                nb_retries += 1
                time_sleep += 5                
                time.sleep(time_sleep)
                continue
            else:
                message = f"Retried 10 times and didn't work, with layers: {layers}, styles: {styles}"
                logger.error(message)
                raise Exception(message) from ex

    ##### Save image to file #####
    # Write image to file...
    if not output_dir.exists():
        output_dir.mkdir()
    with output_filepath.open('wb') as image_file:
        image_file.write(response.read())

    ##### Make the output image compliant with image_format_save #####

    # If geotiff is asked, check if the the coordinates are embedded...
    if image_format_save == FORMAT_GEOTIFF:
        # Read output image to check if coördinates are there
        with rio.open(str(output_filepath)) as image_ds:
            image_profile_orig = image_ds.profile
            image_transform_affine = image_ds.transform

            if image_pixels_ignore_border == 0:
                image_data = image_ds.read()
            else:
                image_data = image_ds.read(window=rio.windows.Window(
                        image_pixels_ignore_border, image_pixels_ignore_border, 
                        size[0], size[1]))

        logger.debug(f"original image_profile: {image_profile_orig}")

        # If coordinates are not embedded add them, if image_pixels_ignore_border
        # change them
        if((image_transform_affine[2] == 0 and image_transform_affine[5] == 0)
            or image_pixels_ignore_border > 0):
            logger.debug(f"Coordinates not present in image, driver: {image_profile_orig['driver']}")

            # If profile format is not gtiff, create new profile
            if image_profile_orig['driver'] != 'GTiff':
                image_profile_gtiff = rio.profiles.DefaultGTiffProfile.defaults

                # Copy appropriate info from source file
                image_profile_gtiff.update(
                        count=image_profile_orig['count'], width=size[0], height=size[1],
                        nodata=image_profile_orig['nodata'], dtype=image_profile_orig['dtype'])
                image_profile = image_profile_gtiff
            else:
                image_profile = image_profile_orig
                
            # Set the asked compression
            image_profile.update(compress=tiff_compress)

            logger.debug(f"Map request bbox: {bbox_for_getmap}")
            logger.debug(f"Map request size: {size_for_getmap}")

            # For some coordinate systems apparently the axis ordered is configured wrong in LibOWS :-(
            srs_pixel_x_size = (bbox[2]-bbox[0])/size[0]
            srs_pixel_y_size = (bbox[1]-bbox[3])/size[1]

            logger.debug(f"Coordinates to put in geotiff:\n" +
                         f"    - x-component of the pixel width, W-E: {srs_pixel_x_size}\n" +
                         f"    - y-component of the pixel width, W-E (0 if image is exactly N up): 0\n" +
                         f"    - top-left x: {bbox[0]}\n" +
                         f"    - x-component of the pixel height, N-S (0 if image is exactly N up): \n" +
                         f"    - y-component of the pixel height, N-S: {srs_pixel_y_size}\n" +
                         f"    - top-left y: {bbox[3]}")

            # Add transform and srs to the profile
            image_profile.update(
                    transform = rio.transform.Affine(
                                srs_pixel_x_size, 0, bbox[0],
                                0 , srs_pixel_y_size, bbox[3]),
                    crs=srs)

            # Delete output file, and write again
            output_filepath.unlink()
            with rio.open(str(output_filepath), 'w', **image_profile) as image_file:
                image_file.write(image_data)

    else:
        # For file formats that doesn't support coordinates, we add a worldfile       
        srs_pixel_x_size = (bbox[2]-bbox[0])/size[0]
        srs_pixel_y_size = (bbox[1]-bbox[3])/size[1]

        path_noext = output_filepath.parent / output_filepath.stem
        ext_world = get_world_ext_for_image_format(image_format_save)
        output_worldfile_filepath = Path(str(path_noext) + ext_world)
        
        with output_worldfile_filepath.open('w') as wld_file:
            wld_file.write(f"{srs_pixel_x_size}")
            wld_file.write("\n0.000")
            wld_file.write("\n0.000")
            wld_file.write(f"\n{srs_pixel_y_size}")
            wld_file.write(f"\n{bbox[0]}")
            wld_file.write(f"\n{bbox[3]}")
        
        # If the image format to save is different, or if a border needs to be ignored
        if(image_format != image_format_save 
           or image_pixels_ignore_border > 0):

            # Read image 
            with rio.open(str(output_filepath)) as image_ds:
                image_profile_orig = image_ds.profile
                image_transform_affine = image_ds.transform

                # If border needs to be ignored, only read data we are interested in
                if image_pixels_ignore_border == 0:
                    image_data = image_ds.read()
                else:
                    image_data = image_ds.read(window=rio.windows.Window(
                            image_pixels_ignore_border, image_pixels_ignore_border, 
                            size[0], size[1]))
            
            # If same save format, reuse profile
            if image_format == image_format_save:
                image_profile_output = image_profile_orig
                if image_pixels_ignore_border != 0:
                    image_profile_output.update(width=size[0], height=size[1])
            else:
                if image_format_save == FORMAT_TIFF:
                    driver = 'GTiff'
                    compress=tiff_compress
                else:
                    raise Exception(f"Unsupported image_format_save: {image_format_save}")
                image_profile_output = rio.profiles.Profile(
                        width=size[0], height=size[1], count=image_profile_orig['count'],
                        nodata=image_profile_orig['nodata'], dtype=image_profile_orig['dtype'],
                        compress=compress, driver=driver)

            # Delete output file, and write again
            output_filepath.unlink()
            image_profile_output = get_cleaned_write_profile(image_profile_output)
            with rio.open(str(output_filepath), 'w', **image_profile_output) as image_file:
                image_file.write(image_data)               

            #raise Exception(f"Different save format not supported between {image_format} and {image_format_save}")

    return output_filepath

def create_filename(
        srs: str,
        bbox,
        size,
        image_format: str,
        layername: str = None):
    
    # Get image extension based on format
    image_ext = get_ext_for_image_format(image_format)

    # Use different file names for projected vs geographic SRS
    is_srs_projected = rio.crs.CRS.from_string(srs).is_projected
    if is_srs_projected:
        output_filename = f"{bbox[0]:06.0f}_{bbox[1]:06.0f}_{bbox[2]:06.0f}_{bbox[3]:06.0f}_{size[0]}_{size[1]}"
    else:
        output_filename = f"{bbox[0]:09.4f}_{bbox[1]:09.4f}_{bbox[2]:09.4f}_{bbox[3]:09.4f}_{size[0]}_{size[1]}"

    # Add layername if it is not None    
    if layername is not None:
        output_filename += "_" + layername
    
    # Add extension
    output_filename += image_ext

    return output_filename

def get_ext_for_image_format(image_format: str) -> str:
    # Choose image extension based on format
    if image_format == FORMAT_GEOTIFF:
        return FORMAT_GEOTIFF_EXT
    elif image_format == FORMAT_TIFF:
        return FORMAT_TIFF_EXT
    elif image_format == FORMAT_JPEG:
        return FORMAT_JPEG_EXT
    elif image_format == FORMAT_PNG:
        return FORMAT_PNG_EXT
    else:
        raise Exception(f"get_ext_for_image_format for image format {image_format} is not implemented!")

def get_world_ext_for_image_format(image_format: str) -> str:
    # Choose image extension based on format
    if image_format == FORMAT_GEOTIFF:
        return FORMAT_GEOTIFF_EXT_WORLD
    elif image_format == FORMAT_TIFF:
        return FORMAT_TIFF_EXT_WORLD
    elif image_format == FORMAT_JPEG:
        return FORMAT_JPEG_EXT_WORLD
    elif image_format == FORMAT_PNG:
        return FORMAT_PNG_EXT_WORLD
    else:
        raise Exception(f"get_world_ext_for_image_format for image format {image_format} is not implemented!")

def get_cleaned_write_profile(profile: dict) -> dict:

    # Depending on the driver, different profile keys are supported    
    if profile['driver'] == 'JPEG':
        # Don't copy profile keys to cleaned version that are not supported for JPEG
        profile_cleaned = {}
        for profile_key in profile:
            if profile_key not in ['tiled', 'compress', 'interleave', 'photometric']:
                profile_cleaned[profile_key] = profile[profile_key]
    elif profile['driver'] == 'PNG':
        # Don't copy profile keys to cleaned version that are not supported for JPEG
        profile_cleaned = {}
        for profile_key in profile:
            if profile_key not in ['tiled', 'interleave']:
                profile_cleaned[profile_key] = profile[profile_key]
    else:
        profile_cleaned = profile.copy()

    return profile_cleaned



# -*- coding: utf-8 -*-
"""
Module to make it easy to start a training session.
"""

import argparse
import os
from pathlib import Path
import shlex
import sys

from orthoseg.helpers import config_helper as conf
from orthoseg.helpers import log_helper
from orthoseg.lib import prepare_traindatasets as prep
from orthoseg.lib import predicter
from orthoseg.lib import trainer
from orthoseg.model import model_factory as mf
from orthoseg.model import model_helper as mh

def train_argstr(argstr):
    args = shlex.split(argstr)
    train_args(args)

def train_args(args):

    ##### Interprete arguments #####
    parser = argparse.ArgumentParser(add_help=False)

    # Required arguments
    required = parser.add_argument_group('Required arguments')
    required.add_argument("--configfile", type=str, required=True,
            help="The project config file to use")
    
    # Optional arguments
    optional = parser.add_argument_group('Optional arguments')
    # Add back help 
    optional.add_argument('-h', '--help', action='help', default=argparse.SUPPRESS,
            help='Show this help message and exit')

    # Interprete arguments
    args = parser.parse_args(args)

    ##### Run! #####
    train(projectconfig_path=Path(args.configfile))

def train(
        projectconfig_path: Path,
        imagelayerconfig_path: Path = None):
    """
    Run a training session.

    Args:
        projectconfig_path (Path): Path to the projects config file.
        imagelayerconfig_path (Path, optional): Path to the imagelayer config file. If not specified, 
            the path specified in files.image_layers_config_filepath in the project config will be used. 
            Defaults to None.
    """
    ##### Init #####   
    # Load config
    config_filepaths = conf.search_projectconfig_files(projectconfig_path=projectconfig_path)
    conf.read_project_config(config_filepaths, imagelayerconfig_path)
    
    # Main initialisation of the logging
    global logger
    logger = log_helper.main_log_init(conf.dirs.getpath('log_training_dir'), __name__)      
    logger.debug(f"Config used: \n{conf.pformat_config()}")
    # TODO: add something to delete old data, predictions???

    # First check if the segment_subject has a valid name
    segment_subject = conf.general['segment_subject']
    if segment_subject == 'MUST_OVERRIDE':
        raise Exception("The segment_subject parameter needs to be overridden in the subject specific config file!!!") 
    elif '_' in segment_subject:
        raise Exception(f"The segment_subject parameter should not contain '_', so this is invalid: {segment_subject}!!!") 

    # Create the output dir's if they don't exist yet...
    for dir in [conf.dirs.getpath('project_dir'), conf.dirs.getpath('training_dir')]:
        if dir and not dir.exists():
            dir.mkdir()
    
    ##### If the training data doesn't exist yet, create it #####
    # Get the label input info, and clean it so it is practical to use
    label_files = conf.train.getdict('label_datasources')
    for label_file_key in label_files:
        # Convert the str file paths to Path objects
        label_files[label_file_key]['locations_path'] = Path(
                label_files[label_file_key]['locations_path'])
        label_files[label_file_key]['data_path'] = Path(
                label_files[label_file_key]['data_path'])
    first_label_file = list(label_files)[0]
    train_image_layer = label_files[first_label_file]['image_layer']
    train_projection = conf.image_layers[train_image_layer]['projection']
    classes = conf.train.getdict('classes')

    # Now create the train datasets (train, validation, test)
    force_model_traindata_id = conf.train.getint('force_model_traindata_id')
    if force_model_traindata_id > -1:
        training_dir = conf.dirs.getpath('training_dir') / f"{force_model_traindata_id:02d}"
        traindata_id = force_model_traindata_id
    else:
        logger.info("Prepare train, validation and test data")
        training_dir, traindata_id = prep.prepare_traindatasets(
                label_files=label_files,
                classes=classes,
                image_layers=conf.image_layers,
                training_dir=conf.dirs.getpath('training_dir'),
                image_pixel_x_size=conf.train.getfloat('image_pixel_x_size'),
                image_pixel_y_size=conf.train.getfloat('image_pixel_y_size'),
                image_pixel_width=conf.train.getint('image_pixel_width'),
                image_pixel_height=conf.train.getint('image_pixel_height'))
    logger.info(f"Traindata dir to use is {training_dir}, with traindata_id: {traindata_id}")
    traindata_dir = training_dir / 'train'
    validationdata_dir = training_dir / 'validation'
    testdata_dir = training_dir / 'test'
    
    ##### Check if training is needed #####
    # Get hyper parameters from the config
    architectureparams = mh.ArchitectureParams(
            architecture=conf.model['architecture'],
            nb_classes=len(classes),   
            nb_channels=conf.model.getint('nb_channels'),
            architecture_id=conf.model.getint('architecture_id'))
    trainparams = mh.TrainParams(
            trainparams_id=conf.train.getint('trainparams_id'),
            image_augmentations=conf.train.getdict('image_augmentations'),
            mask_augmentations=conf.train.getdict('mask_augmentations'),
            class_weights=[classes[class_name]['weight'] for class_name in classes],
            batch_size=conf.train.getint('batch_size_fit'), 
            optimizer=conf.train.get('optimizer'), 
            optimizer_params=conf.train.getdict('optimizer_params'), 
            loss_function=conf.train.get('loss_function'), 
            monitor_metric=conf.train.get('monitor_metric'), 
            monitor_metric_mode=conf.train.get('monitor_metric_mode'), 
            save_format=conf.train.get('save_format'), 
            save_best_only=conf.train.getboolean('save_best_only'), 
            save_min_accuracy=conf.train.getfloat('save_min_accuracy'),
            nb_epoch=conf.train.getint('max_epoch'),
            earlystop_patience=conf.train.getint('earlystop_patience'),
            earlystop_monitor_metric=conf.train.get('earlystop_monitor_metric'))

    # Check if there exists already a model for this train dataset + hyperparameters
    model_dir = conf.dirs.getpath('model_dir')
    segment_subject = conf.general['segment_subject']
    best_model_curr_train_version = mh.get_best_model(
            model_dir=model_dir, 
            segment_subject=segment_subject,
            traindata_id=traindata_id,
            architecture_id=architectureparams.architecture_id,
            trainparams_id=trainparams.trainparams_id)

    # Determine if training is needed,...
    resume_train = conf.train.getboolean('resume_train')
    if resume_train is False:
        # If no (best) model found, training needed!
        if best_model_curr_train_version is None:
            train_needed = True
        elif conf.train.getboolean('force_train') is True:
            train_needed = True
        else:
            logger.info("JUST PREDICT, without training: preload_existing_model is false and model found")
            train_needed = False
    else:
        # We want to preload an existing model and models were found
        if best_model_curr_train_version is not None:
            logger.info(f"PRELOAD model and continue TRAINING it: {best_model_curr_train_version['filename']}")
            train_needed = True
        else:
            message = "STOP: preload_existing_model is true but no model was found!"
            logger.error(message)
            raise Exception(message)
    
    ##### Train!!! #####
    if train_needed is True:

        # If a model already exists, use it to predict (possibly new) training and 
        # validation dataset. This way it is possible to have a quick check on errors
        # in (new) added labels in the datasets.

        # Get the current best model that already exists for this subject
        best_recent_model = mh.get_best_model(model_dir=model_dir)
        if best_recent_model is not None:
            try:
                logger.info(f"Load model + weights from {best_recent_model['filepath']}")    
                model = mf.load_model(best_recent_model['filepath'], compile=False)            
                logger.info("Loaded model + weights")

                # Prepare output subdir to be used for predictions
                predict_out_subdir, _ = os.path.splitext(best_recent_model['filename'])
                
                # Predict training dataset
                predicter.predict_dir(
                        model=model,
                        input_image_dir=traindata_dir / 'image',
                        output_base_dir=traindata_dir / predict_out_subdir,
                        projection_if_missing=train_projection,
                        input_mask_dir=traindata_dir / 'mask',
                        batch_size=conf.train.getint('batch_size_predict'), 
                        evaluate_mode=True)
                    
                # Predict validation dataset
                predicter.predict_dir(
                        model=model,
                        input_image_dir=validationdata_dir / 'image',
                        output_base_dir=validationdata_dir / predict_out_subdir,
                        projection_if_missing=train_projection,
                        input_mask_dir=validationdata_dir / 'mask',
                        batch_size=conf.train.getint('batch_size_predict'), 
                        evaluate_mode=True)
                del model
            except Exception as ex:
                logger.warn(f"Exception trying to predict with old model: {ex}")
        
        # Now we can really start training
        logger.info('Start training')
        model_preload_filepath = None
        if best_model_curr_train_version is not None:
            model_preload_filepath = best_model_curr_train_version['filepath']
        elif conf.train.getboolean('preload_with_previous_traindata'):
            best_model_for_architecture = mh.get_best_model(
                    model_dir=model_dir, 
                    segment_subject=segment_subject)
            if best_model_for_architecture is not None:
                model_preload_filepath = best_model_for_architecture['filepath']
        
        # Combine all hyperparameters in hyperparams object
        hyperparams = mh.HyperParams(
                architecture=architectureparams,
                train=trainparams)

        trainer.train(
                traindata_dir=traindata_dir,
                validationdata_dir=validationdata_dir,
                model_save_dir=model_dir,
                segment_subject=segment_subject,
                traindata_id=traindata_id,
                hyperparams=hyperparams,
                model_preload_filepath=model_preload_filepath,
                image_width=conf.train.getint('image_pixel_width'),
                image_height=conf.train.getint('image_pixel_height')) 
    
        # Now get the best model found during training
        best_model_curr_train_version = mh.get_best_model(
                model_dir=model_dir, 
                segment_subject=segment_subject,
                traindata_id=traindata_id)

    # Now predict on the train,... data  
    logger.info(f"PREDICT test data with best model: {best_model_curr_train_version['filename']}")
    
    # Load prediction model...
    logger.info(f"Load model + weights from {best_model_curr_train_version['filepath']}")    
    model = mf.load_model(best_model_curr_train_version['filepath'], compile=False)            
    logger.info("Loaded model + weights")
    
    # Prepare output subdir to be used for predictions
    predict_out_subdir, _ = os.path.splitext(best_model_curr_train_version['filename'])
    
    # Predict training dataset
    predicter.predict_dir(
            model=model,
            input_image_dir=traindata_dir / 'image',
            output_base_dir=traindata_dir / predict_out_subdir,
            projection_if_missing=train_projection,
            input_mask_dir=traindata_dir / 'mask',
            batch_size=conf.train.getint('batch_size_predict'), 
            evaluate_mode=True)
    
    # Predict validation dataset
    predicter.predict_dir(
            model=model,
            input_image_dir=validationdata_dir / 'image',
            output_base_dir=validationdata_dir / predict_out_subdir,
            projection_if_missing=train_projection,
            input_mask_dir=validationdata_dir / 'mask',
            batch_size=conf.train.getint('batch_size_predict'), 
            evaluate_mode=True)

    # Predict test dataset, if it exists
    if testdata_dir is not None and testdata_dir.exists():
        predicter.predict_dir(
                model=model,
                input_image_dir=testdata_dir / 'image',
                output_base_dir=testdata_dir / predict_out_subdir, 
                projection_if_missing=train_projection,
                input_mask_dir=testdata_dir / 'mask',
                batch_size=conf.train.getint('batch_size_predict'), 
                evaluate_mode=True)
    
    # Predict extra test dataset with random images in the roi, to add to 
    # train and/or validation dataset if inaccuracies are found
    # -> this is very useful to find false positives to improve the datasets
    if conf.dirs.getpath('predictsample_image_input_dir').exists():
        predicter.predict_dir(
                model=model,
                input_image_dir=conf.dirs.getpath('predictsample_image_input_dir'),
                output_base_dir=conf.dirs.getpath('predictsample_image_output_basedir') / predict_out_subdir,
                projection_if_missing=train_projection,
                batch_size=conf.train.getint('batch_size_predict'), 
                evaluate_mode=True)
    
if __name__ == '__main__':
    train_args(sys.argv[1:])
    


# -*- coding: utf-8 -*-
""" Deeplabv3+ model for Keras.

Downloaded from: https://github.com/bonlime/keras-deeplab-v3-plus/blob/master/model.py

This model is based on TF repo:
https://github.com/tensorflow/models/tree/master/research/deeplab
On Pascal VOC, original model gets to 84.56% mIOU

MobileNetv2 backbone is based on this repo:
https://github.com/JonathanCMitchell/mobilenet_v2_keras

# Reference
- [Encoder-Decoder with Atrous Separable Convolution
    for Semantic Image Segmentation](https://arxiv.org/pdf/1802.02611.pdf)
- [Xception: Deep Learning with Depthwise Separable Convolutions]
    (https://arxiv.org/abs/1610.02357)
- [Inverted Residuals and Linear Bottlenecks: Mobile Networks for
    Classification, Detection and Segmentation](https://arxiv.org/abs/1801.04381)
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

'''
from tensorflow.keras import layers
from tensorflow.keras.layers import Input
from tensorflow.keras.layers import Lambda
from tensorflow.keras.layers import Activation
from tensorflow.keras.layers import Concatenate
from tensorflow.keras.layers import Add
from tensorflow.keras.layers import Dropout
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import DepthwiseConv2D
from tensorflow.keras.layers import ZeroPadding2D
from tensorflow.keras.layers import GlobalAveragePooling2D
from tensorflow.keras.utils.layer_utils import get_source_inputs
from tensorflow.keras.utils.data_utils import get_file
from tensorflow.keras import backend as K
from tensorflow.keras.applications.imagenet_utils import preprocess_input
'''

WEIGHTS_PATH_X = "https://github.com/bonlime/keras-deeplab-v3-plus/releases/download/1.1/deeplabv3_xception_tf_dim_ordering_tf_kernels.h5"
WEIGHTS_PATH_MOBILE = "https://github.com/bonlime/keras-deeplab-v3-plus/releases/download/1.1/deeplabv3_mobilenetv2_tf_dim_ordering_tf_kernels.h5"
WEIGHTS_PATH_X_CS = "https://github.com/bonlime/keras-deeplab-v3-plus/releases/download/1.2/deeplabv3_xception_tf_dim_ordering_tf_kernels_cityscapes.h5"
WEIGHTS_PATH_MOBILE_CS = "https://github.com/bonlime/keras-deeplab-v3-plus/releases/download/1.2/deeplabv3_mobilenetv2_tf_dim_ordering_tf_kernels_cityscapes.h5"

def SepConv_BN(x, filters, prefix, stride=1, kernel_size=3, rate=1, depth_activation=False, epsilon=1e-3):
    """ SepConv with BN between depthwise & pointwise. Optionally add activation after BN
        Implements right "same" padding for even kernel sizes
        Args:
            x: input tensor
            filters: num of filters in pointwise convolution
            prefix: prefix before name
            stride: stride at depthwise conv
            kernel_size: kernel size for depthwise convolution
            rate: atrous rate for depthwise convolution
            depth_activation: flag to use activation between depthwise & poinwise convs
            epsilon: epsilon to use in BN layer
    """

    if stride == 1:
        depth_padding = 'same'
    else:
        kernel_size_effective = kernel_size + (kernel_size - 1) * (rate - 1)
        pad_total = kernel_size_effective - 1
        pad_beg = pad_total // 2
        pad_end = pad_total - pad_beg
        x = tf.keras.layers.ZeroPadding2D((pad_beg, pad_end))(x)
        depth_padding = 'valid'

    if not depth_activation:
        x = tf.keras.layers.Activation(tf.nn.relu)(x)
    x = tf.keras.layers.DepthwiseConv2D((kernel_size, kernel_size), strides=(stride, stride), dilation_rate=(rate, rate),
                        padding=depth_padding, use_bias=False, name=prefix + '_depthwise')(x)
    x = tf.keras.layers.BatchNormalization(name=prefix + '_depthwise_BN', epsilon=epsilon)(x)
    if depth_activation:
        x = tf.keras.layers.Activation(tf.nn.relu)(x)
    x = tf.keras.layers.Conv2D(filters, (1, 1), padding='same',
               use_bias=False, name=prefix + '_pointwise')(x)
    x = tf.keras.layers.BatchNormalization(name=prefix + '_pointwise_BN', epsilon=epsilon)(x)
    if depth_activation:
        x = tf.keras.layers.Activation(tf.nn.relu)(x)

    return x


def _conv2d_same(x, filters, prefix, stride=1, kernel_size=3, rate=1):
    """Implements right 'same' padding for even kernel sizes
        Without this there is a 1 pixel drift when stride = 2
        Args:
            x: input tensor
            filters: num of filters in pointwise convolution
            prefix: prefix before name
            stride: stride at depthwise conv
            kernel_size: kernel size for depthwise convolution
            rate: atrous rate for depthwise convolution
    """
    if stride == 1:
        return tf.keras.layers.Conv2D(filters,
                      (kernel_size, kernel_size),
                      strides=(stride, stride),
                      padding='same', use_bias=False,
                      dilation_rate=(rate, rate),
                      name=prefix)(x)
    else:
        kernel_size_effective = kernel_size + (kernel_size - 1) * (rate - 1)
        pad_total = kernel_size_effective - 1
        pad_beg = pad_total // 2
        pad_end = pad_total - pad_beg
        x = tf.keras.layers.ZeroPadding2D((pad_beg, pad_end))(x)
        return tf.keras.layers.Conv2D(filters,
                      (kernel_size, kernel_size),
                      strides=(stride, stride),
                      padding='valid', use_bias=False,
                      dilation_rate=(rate, rate),
                      name=prefix)(x)


def _xception_block(inputs, depth_list, prefix, skip_connection_type, stride,
                    rate=1, depth_activation=False, return_skip=False):
    """ Basic building block of modified Xception network
        Args:
            inputs: input tensor
            depth_list: number of filters in each SepConv layer. len(depth_list) == 3
            prefix: prefix before name
            skip_connection_type: one of {'conv','sum','none'}
            stride: stride at last depthwise conv
            rate: atrous rate for depthwise convolution
            depth_activation: flag to use activation between depthwise & pointwise convs
            return_skip: flag to return additional tensor after 2 SepConvs for decoder
            """
    residual = inputs
    for i in range(3):
        residual = SepConv_BN(residual,
                              depth_list[i],
                              prefix + '_separable_conv{}'.format(i + 1),
                              stride=stride if i == 2 else 1,
                              rate=rate,
                              depth_activation=depth_activation)
        if i == 1:
            skip = residual
    if skip_connection_type == 'conv':
        shortcut = _conv2d_same(inputs, depth_list[-1], prefix + '_shortcut',
                                kernel_size=1,
                                stride=stride)
        shortcut = tf.keras.layers.BatchNormalization(name=prefix + '_shortcut_BN')(shortcut)
        outputs = tf.keras.layers.add([residual, shortcut])
    elif skip_connection_type == 'sum':
        outputs = tf.keras.layers.add([residual, inputs])
    elif skip_connection_type == 'none':
        outputs = residual
    if return_skip:
        return outputs, skip
    else:
        return outputs


def _make_divisible(v, divisor, min_value=None):
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v


def _inverted_res_block(inputs, expansion, stride, alpha, filters, block_id, skip_connection, rate=1):
    in_channels = inputs.shape[-1].value  # inputs._keras_shape[-1]
    pointwise_conv_filters = int(filters * alpha)
    pointwise_filters = _make_divisible(pointwise_conv_filters, 8)
    x = inputs
    prefix = 'expanded_conv_{}_'.format(block_id)
    if block_id:
        # Expand

        x = tf.keras.layers.Conv2D(expansion * in_channels, kernel_size=1, padding='same',
                   use_bias=False, activation=None,
                   name=prefix + 'expand')(x)
        x = tf.keras.layers.BatchNormalization(epsilon=1e-3, momentum=0.999,
                               name=prefix + 'expand_BN')(x)
        x = tf.keras.layers.Activation(tf.nn.relu6, name=prefix + 'expand_relu')(x)
    else:
        prefix = 'expanded_conv_'
    # Depthwise
    x = tf.keras.layers.DepthwiseConv2D(kernel_size=3, strides=stride, activation=None,
                        use_bias=False, padding='same', dilation_rate=(rate, rate),
                        name=prefix + 'depthwise')(x)
    x = tf.keras.layers.BatchNormalization(epsilon=1e-3, momentum=0.999,
                           name=prefix + 'depthwise_BN')(x)

    x = tf.keras.layers.Activation(tf.nn.relu6, name=prefix + 'depthwise_relu')(x)

    # Project
    x = tf.keras.layers.Conv2D(pointwise_filters,
               kernel_size=1, padding='same', use_bias=False, activation=None,
               name=prefix + 'project')(x)
    x = tf.keras.layers.BatchNormalization(epsilon=1e-3, momentum=0.999,
                           name=prefix + 'project_BN')(x)

    if skip_connection:
        return tf.keras.layers.Add(name=prefix + 'add')([inputs, x])

    # if in_channels == pointwise_filters and stride == 1:
    #    return tf.keras.layers.Add(name='res_connect_' + str(block_id))([inputs, x])

    return x


def Deeplabv3(weights='pascal_voc', input_tensor=None, input_shape=(512, 512, 3), classes=21, backbone='mobilenetv2',
              OS=16, alpha=1., activation=None):
    """ Instantiates the Deeplabv3+ architecture

    Optionally loads weights pre-trained
    on PASCAL VOC or Cityscapes. This model is available for TensorFlow only.
    # Arguments
        weights: one of 'pascal_voc' (pre-trained on pascal voc),
            'cityscapes' (pre-trained on cityscape) or None (random initialization)
        input_tensor: optional Keras tensor (i.e. output of `layers.Input()`)
            to use as image input for the model.
        input_shape: shape of input image. format HxWxC
            PASCAL VOC model was trained on (512,512,3) images. None is allowed as shape/width
        classes: number of desired classes. PASCAL VOC has 21 classes, Cityscapes has 19 classes.
            If number of classes not aligned with the weights used, last layer is initialized randomly
        backbone: backbone to use. one of {'xception','mobilenetv2'}
        activation: optional activation to add to the top of the network.
            One of 'softmax', 'sigmoid' or None
        OS: determines input_shape/feature_extractor_output ratio. One of {8,16}.
            Used only for xception backbone.
        alpha: controls the width of the MobileNetV2 network. This is known as the
            width multiplier in the MobileNetV2 paper.
                - If `alpha` < 1.0, proportionally decreases the number
                    of filters in each layer.
                - If `alpha` > 1.0, proportionally increases the number
                    of filters in each layer.
                - If `alpha` = 1, default number of filters from the paper
                    are used at each layer.
            Used only for mobilenetv2 backbone. Pretrained is only available for alpha=1.

    # Returns
        A Keras model instance.

    # Raises
        RuntimeError: If attempting to run this model with a
            backend that does not support separable convolutions.
        ValueError: in case of invalid argument for `weights` or `backbone`

    """

    if not (weights in {'pascal_voc', 'cityscapes', None}):
        raise ValueError('The `weights` argument should be either '
                         '`None` (random initialization), `pascal_voc`, or `cityscapes` '
                         '(pre-trained on PASCAL VOC)')

    if not (backbone in {'xception', 'mobilenetv2'}):
        raise ValueError('The `backbone` argument should be either '
                         '`xception`  or `mobilenetv2` ')

    if input_tensor is None:
        img_input = tf.keras.layers.Input(shape=input_shape)
    else:
        img_input = input_tensor

    if backbone == 'xception':
        if OS == 8:
            entry_block3_stride = 1
            middle_block_rate = 2  # ! Not mentioned in paper, but required
            exit_block_rates = (2, 4)
            atrous_rates = (12, 24, 36)
        else:
            entry_block3_stride = 2
            middle_block_rate = 1
            exit_block_rates = (1, 2)
            atrous_rates = (6, 12, 18)

        x = tf.keras.layers.Conv2D(32, (3, 3), strides=(2, 2),
                   name='entry_flow_conv1_1', use_bias=False, padding='same')(img_input)
        x = tf.keras.layers.BatchNormalization(name='entry_flow_conv1_1_BN')(x)
        x = tf.keras.layers.Activation(tf.nn.relu)(x)

        x = _conv2d_same(x, 64, 'entry_flow_conv1_2', kernel_size=3, stride=1)
        x = tf.keras.layers.BatchNormalization(name='entry_flow_conv1_2_BN')(x)
        x = tf.keras.layers.Activation(tf.nn.relu)(x)

        x = _xception_block(x, [128, 128, 128], 'entry_flow_block1',
                            skip_connection_type='conv', stride=2,
                            depth_activation=False)
        x, skip1 = _xception_block(x, [256, 256, 256], 'entry_flow_block2',
                                   skip_connection_type='conv', stride=2,
                                   depth_activation=False, return_skip=True)

        x = _xception_block(x, [728, 728, 728], 'entry_flow_block3',
                            skip_connection_type='conv', stride=entry_block3_stride,
                            depth_activation=False)
        for i in range(16):
            x = _xception_block(x, [728, 728, 728], 'middle_flow_unit_{}'.format(i + 1),
                                skip_connection_type='sum', stride=1, rate=middle_block_rate,
                                depth_activation=False)

        x = _xception_block(x, [728, 1024, 1024], 'exit_flow_block1',
                            skip_connection_type='conv', stride=1, rate=exit_block_rates[0],
                            depth_activation=False)
        x = _xception_block(x, [1536, 1536, 2048], 'exit_flow_block2',
                            skip_connection_type='none', stride=1, rate=exit_block_rates[1],
                            depth_activation=True)

    else:
        OS = 8
        first_block_filters = _make_divisible(32 * alpha, 8)
        x = tf.keras.layers.Conv2D(first_block_filters,
                   kernel_size=3,
                   strides=(2, 2), padding='same',
                   use_bias=False, name='Conv')(img_input)
        x = tf.keras.layers.BatchNormalization(
            epsilon=1e-3, momentum=0.999, name='Conv_BN')(x)
        x = tf.keras.layers.Activation(tf.nn.relu6, name='Conv_Relu6')(x)

        x = _inverted_res_block(x, filters=16, alpha=alpha, stride=1,
                                expansion=1, block_id=0, skip_connection=False)

        x = _inverted_res_block(x, filters=24, alpha=alpha, stride=2,
                                expansion=6, block_id=1, skip_connection=False)
        x = _inverted_res_block(x, filters=24, alpha=alpha, stride=1,
                                expansion=6, block_id=2, skip_connection=True)

        x = _inverted_res_block(x, filters=32, alpha=alpha, stride=2,
                                expansion=6, block_id=3, skip_connection=False)
        x = _inverted_res_block(x, filters=32, alpha=alpha, stride=1,
                                expansion=6, block_id=4, skip_connection=True)
        x = _inverted_res_block(x, filters=32, alpha=alpha, stride=1,
                                expansion=6, block_id=5, skip_connection=True)

        # stride in block 6 changed from 2 -> 1, so we need to use rate = 2
        x = _inverted_res_block(x, filters=64, alpha=alpha, stride=1,  # 1!
                                expansion=6, block_id=6, skip_connection=False)
        x = _inverted_res_block(x, filters=64, alpha=alpha, stride=1, rate=2,
                                expansion=6, block_id=7, skip_connection=True)
        x = _inverted_res_block(x, filters=64, alpha=alpha, stride=1, rate=2,
                                expansion=6, block_id=8, skip_connection=True)
        x = _inverted_res_block(x, filters=64, alpha=alpha, stride=1, rate=2,
                                expansion=6, block_id=9, skip_connection=True)

        x = _inverted_res_block(x, filters=96, alpha=alpha, stride=1, rate=2,
                                expansion=6, block_id=10, skip_connection=False)
        x = _inverted_res_block(x, filters=96, alpha=alpha, stride=1, rate=2,
                                expansion=6, block_id=11, skip_connection=True)
        x = _inverted_res_block(x, filters=96, alpha=alpha, stride=1, rate=2,
                                expansion=6, block_id=12, skip_connection=True)

        x = _inverted_res_block(x, filters=160, alpha=alpha, stride=1, rate=2,  # 1!
                                expansion=6, block_id=13, skip_connection=False)
        x = _inverted_res_block(x, filters=160, alpha=alpha, stride=1, rate=4,
                                expansion=6, block_id=14, skip_connection=True)
        x = _inverted_res_block(x, filters=160, alpha=alpha, stride=1, rate=4,
                                expansion=6, block_id=15, skip_connection=True)

        x = _inverted_res_block(x, filters=320, alpha=alpha, stride=1, rate=4,
                                expansion=6, block_id=16, skip_connection=False)

    # end of feature extractor

    # branching for Atrous Spatial Pyramid Pooling

    # Image Feature branch
    shape_before = tf.shape(x)
    b4 = tf.keras.layers.GlobalAveragePooling2D()(x)
    # from (b_size, channels)->(b_size, 1, 1, channels)
    b4 = tf.keras.layers.Lambda(lambda x: tf.keras.backend.expand_dims(x, 1))(b4)
    b4 = tf.keras.layers.Lambda(lambda x: tf.keras.backend.expand_dims(x, 1))(b4)
    b4 = tf.keras.layers.Conv2D(256, (1, 1), padding='same',
                use_bias=False, name='image_pooling')(b4)
    b4 = tf.keras.layers.BatchNormalization(name='image_pooling_BN', epsilon=1e-5)(b4)
    b4 = tf.keras.layers.Activation(tf.nn.relu)(b4)
    # upsample. have to use compat because of the option align_corners
    size_before = tf.keras.backend.int_shape(x)
    b4 = tf.keras.layers.Lambda(lambda x: tf.compat.v1.image.resize(x, size_before[1:3],
                                                    method='bilinear', align_corners=True))(b4)
    # simple 1x1
    b0 = tf.keras.layers.Conv2D(256, (1, 1), padding='same', use_bias=False, name='aspp0')(x)
    b0 = tf.keras.layers.BatchNormalization(name='aspp0_BN', epsilon=1e-5)(b0)
    b0 = tf.keras.layers.Activation(tf.nn.relu, name='aspp0_activation')(b0)

    # there are only 2 branches in mobilenetV2. not sure why
    if backbone == 'xception':
        # rate = 6 (12)
        b1 = SepConv_BN(x, 256, 'aspp1',
                        rate=atrous_rates[0], depth_activation=True, epsilon=1e-5)
        # rate = 12 (24)
        b2 = SepConv_BN(x, 256, 'aspp2',
                        rate=atrous_rates[1], depth_activation=True, epsilon=1e-5)
        # rate = 18 (36)
        b3 = SepConv_BN(x, 256, 'aspp3',
                        rate=atrous_rates[2], depth_activation=True, epsilon=1e-5)

        # concatenate ASPP branches & project
        x = tf.keras.layers.Concatenate()([b4, b0, b1, b2, b3])
    else:
        x = tf.keras.layers.Concatenate()([b4, b0])

    x = tf.keras.layers.Conv2D(256, (1, 1), padding='same',
               use_bias=False, name='concat_projection')(x)
    x = tf.keras.layers.BatchNormalization(name='concat_projection_BN', epsilon=1e-5)(x)
    x = tf.keras.layers.Activation(tf.nn.relu)(x)
    x = tf.keras.layers.Dropout(0.1)(x)
    # DeepLab v.3+ decoder

    if backbone == 'xception':
        # Feature projection
        # x4 (x2) block
        size_before2 = tf.keras.backend.int_shape(x)
        x = tf.keras.layers.Lambda(lambda xx: tf.compat.v1.image.resize(xx,
                                                        skip1.shape[1:3],
                                                        method='bilinear', align_corners=True))(x)

        dec_skip1 = tf.keras.layers.Conv2D(48, (1, 1), padding='same',
                           use_bias=False, name='feature_projection0')(skip1)
        dec_skip1 = tf.keras.layers.BatchNormalization(
            name='feature_projection0_BN', epsilon=1e-5)(dec_skip1)
        dec_skip1 = tf.keras.layers.Activation(tf.nn.relu)(dec_skip1)
        x = tf.keras.layers.Concatenate()([x, dec_skip1])
        x = SepConv_BN(x, 256, 'decoder_conv0',
                       depth_activation=True, epsilon=1e-5)
        x = SepConv_BN(x, 256, 'decoder_conv1',
                       depth_activation=True, epsilon=1e-5)

    # you can use it with arbitary number of classes
    if (weights == 'pascal_voc' and classes == 21) or (weights == 'cityscapes' and classes == 19):
        last_layer_name = 'logits_semantic'
    else:
        last_layer_name = 'custom_logits_semantic'

    x = tf.keras.layers.Conv2D(classes, (1, 1), padding='same', name=last_layer_name)(x)
    size_before3 = tf.keras.backend.int_shape(img_input)
    x = tf.keras.layers.Lambda(lambda xx: tf.compat.v1.image.resize(xx,
                                                    size_before3[1:3],
                                                    method='bilinear', align_corners=True))(x)

    # Ensure that the model takes into account
    # any potential predecessors of `input_tensor`.
    if input_tensor is not None:
        inputs = tf.keras.utils.layer_utils.get_source_inputs(input_tensor)
    else:
        inputs = img_input

    if activation in {'softmax', 'sigmoid'}:
        x = tf.keras.layers.Activation(activation)(x)

    model = tf.keras.Model(inputs, x, name='deeplabv3plus')

    # load weights

    if weights == 'pascal_voc':
        if backbone == 'xception':
            weights_path = tf.keras.utils.layer_utils.get_file('deeplabv3_xception_tf_dim_ordering_tf_kernels.h5',
                                    WEIGHTS_PATH_X,
                                    cache_subdir='models')
        else:
            weights_path = tf.keras.utils.layer_utils.get_file('deeplabv3_mobilenetv2_tf_dim_ordering_tf_kernels.h5',
                                    WEIGHTS_PATH_MOBILE,
                                    cache_subdir='models')
        model.load_weights(weights_path, by_name=True)
    elif weights == 'cityscapes':
        if backbone == 'xception':
            weights_path = tf.keras.utils.layer_utils.get_file('deeplabv3_xception_tf_dim_ordering_tf_kernels_cityscapes.h5',
                                    WEIGHTS_PATH_X_CS,
                                    cache_subdir='models')
        else:
            weights_path = tf.keras.utils.layer_utils.get_file('deeplabv3_mobilenetv2_tf_dim_ordering_tf_kernels_cityscapes.h5',
                                    WEIGHTS_PATH_MOBILE_CS,
                                    cache_subdir='models')
        model.load_weights(weights_path, by_name=True)
    return model

def preprocess_input(x):
    """Preprocesses a numpy array encoding a batch of images.
    # Arguments
        x: a 4D numpy array consists of RGB values within [0, 255].
    # Returns
        Input array scaled to [-1.,1.]
    """
    return tf.keras.applications.imagenet_utils.preprocess_input(x, mode='tf')


def get_model(input_width=256, input_height=256, nb_channels=3,
              nb_classes=1,
              loss_mode='binary_crossentropy', learning_rate=0.0001,
              init_model_weights: bool = False, 
              pretrained_weights_filepath: str = None):
    
    weights_to_use = None
    if init_model_weights:
        weights_to_use = 'pascal_voc'
    
    model = Deeplabv3(weights=weights_to_use, 
                     input_tensor=None, 
                     input_shape=(input_width, input_height, nb_channels), 
                     classes=nb_classes, 
                     backbone='xception', 
                     OS=16, 
                     alpha=1.)
    
    model.compile(optimizer=tf.keras.optimizers.Adam(lr=learning_rate), loss=loss_mode,
                  metrics=[jaccard_coef, jacard_coef_flat,
                           jaccard_coef_int, dice_coef, 'accuracy'])
    
    return model


#------------------------------------------
# Metrics functions
#------------------------------------------

SMOOTH_LOSS = 1e-12

def jaccard_coef(y_true, y_pred):
    intersection = tf.keras.backend.sum(y_true * y_pred, axis=[0, -1, -2])
    sum_ = tf.keras.backend.sum(y_true + y_pred, axis=[0, -1, -2])

    jac = (intersection + SMOOTH_LOSS) / (sum_ - intersection + SMOOTH_LOSS)

    return tf.keras.backend.mean(jac)

def jaccard_coef_int(y_true, y_pred):
    y_pred_pos = tf.keras.backend.round(tf.keras.backend.clip(y_pred, 0, 1))

    intersection = tf.keras.backend.sum(y_true * y_pred_pos, axis=[0, -1, -2])
    sum_ = tf.keras.backend.sum(y_true + y_pred_pos, axis=[0, -1, -2])
    jac = (intersection + SMOOTH_LOSS) / (sum_ - intersection + SMOOTH_LOSS)
    return tf.keras.backend.mean(jac)

def jacard_coef_flat(y_true, y_pred):
    y_true_f = tf.keras.backend.flatten(y_true)
    y_pred_f = tf.keras.backend.flatten(y_pred)
    intersection = tf.keras.backend.sum(y_true_f * y_pred_f)
    return (intersection + SMOOTH_LOSS) / (tf.keras.backend.sum(y_true_f) + tf.keras.backend.sum(y_pred_f) - intersection + SMOOTH_LOSS)

def dice_coef(y_true, y_pred, smooth=1.0):
    y_true_f = tf.keras.backend.flatten(y_true)
    y_pred_f = tf.keras.backend.flatten(y_pred)
    intersection = tf.keras.backend.sum(y_true_f * y_pred_f)
    return (2. * intersection + smooth) / (tf.keras.backend.sum(y_true_f) + tf.keras.backend.sum(y_pred_f) + smooth)

def pct_wrong(y_true, y_pred):
    y_pred_pos = tf.keras.backend.round(tf.keras.backend.clip(y_pred, 0, 1))

    intersection = tf.keras.backend.sum(y_true * y_pred_pos, axis=[0, -1, -2])
    sum_ = tf.keras.backend.sum(y_true + y_pred_pos, axis=[0, -1, -2])
    jac = (intersection + SMOOTH_LOSS) / (sum_ - intersection + SMOOTH_LOSS)
    return tf.keras.backend.mean(jac)
